"""Retrieve GeVToMIP calibration constants from Calibration.txt file."""

from __future__ import print_function
from __future__ import absolute_import
import sys
from Helper_Functions import find_between


Calibration_File_And_Path = sys.argv[1]
Energy_To_Calibrate = sys.argv[2]
HCal_Or_ECal_Or_Muon = sys.argv[3]

GeVToMIP = ''

with open(Calibration_File_And_Path, 'r') as f:
    searchlines = f.readlines()
    for line in searchlines:
        if HCal_Or_ECal_Or_Muon == 'HCal':
            if 'HCalGeVToMIP' in line:
                GeVToMIP = find_between(line, ' : ', ' :')
        elif HCal_Or_ECal_Or_Muon == 'ECal':
            if 'ECalGeVToMIP' in line:
                GeVToMIP = find_between(line, ' : ', ' :')
        elif HCal_Or_ECal_Or_Muon == 'Muon':
            if 'MuonGeVToMIP' in line:
                GeVToMIP = find_between(line, ' : ', ' :')
        else:
            print('Please select HCal, ECal or Muon')

Calibration_Text = '_____________________________________________________________________________________' + '\n'

Calibration_Text += 'Extract_GeVToMIP.py extracting ' + HCal_Or_ECal_Or_Muon + 'GeVToMIP.' + '\n'

Calibration_Text += 'For Muon events with energy                        :' + str(Energy_To_Calibrate) + ' /GeV \n\n'

Calibration_Text += HCal_Or_ECal_Or_Muon + 'GeVToMIP is:                                   :' + GeVToMIP + ' \n\n'

with open(Calibration_File_And_Path, 'a') as myfile:
    myfile.write(Calibration_Text)

print(str(GeVToMIP))



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test the dirac-ilc-make-productions script."""

from __future__ import absolute_import
import os
import importlib
import six.moves.configparser
from collections import defaultdict

import pytest
from mock import MagicMock as Mock, patch

from DIRAC import S_OK, S_ERROR

from ILCDIRAC.ILCTransformationSystem.Utilities.Utilities import Task

# pylint: disable=protected-access, invalid-name, missing-docstring, redefined-outer-name
THE_SCRIPT = "ILCDIRAC.ILCTransformationSystem.scripts.dirac_clic_make_productions"
theScript = importlib.import_module(THE_SCRIPT)
SCP = "ILCDIRAC.ILCTransformationSystem.scripts.dirac_clic_make_productions.six.moves.configparser.ConfigParser"
__RCSID__ = "$Id$"

CONFIG_DICT = {}


def test_environ():
  """Ensure we are not running optimized python.

  Some assertion in the pytest code are otherwise ignored.
  """
  assert "PYTHONOPTIMIZE" not in os.environ


def configDict():
  """Return dictionary for dirac configuration system."""
  return {'prodGroup': "myProdGroup",
          'detectorModel': 'myDetectorModel',
          'softwareVersion': 'mySoftwareVersion',
          'configVersion': 'my',
          'configPackage': 'ClicConfig',
          'processes': 'process1, process2',
          'energies': '100, 200',
          'eventsPerJobs': '1000, 2000',
          'productionloglevel': 'DEBUGLEVEL3',
          'outputSE': 'CERN-CASTOR',
          'finalOutputSE': 'VAULT-101',
          'additionalName': 'waitForIt',
          'prodIDs': '123, 456',
          'eventsInSplitFiles': '5000, 6000',
          'ProdTypes': 'Gen, RecOver',
          'MoveTypes': '',
          'MoveStatus': 'Active',
          'MoveGroupSize': '11',
          'overlayEvents': '',
          'overlayEventType': '',
          'cliReco': '--Config.Tracking=Tracked',
          'whizard2Version': 'myWhizardVersion',
          'whizard2SinFile': 'myWhizardSinFile1, myWhizardSinFile2',
          'numberOfTasks': '1, 2',
          'ignoreMetadata': '',
          'taskNames': 'taskA, taskB',
          }


@pytest.fixture
def opsMock():
  """Return fixture for Operations."""
  def mockOpsConfig(*args, **kwargs):  # pylint: disable=unused-argument
    """Mock the operations getValue calls."""
    opsDict = {'DefaultDetectorModel': 'detModel',
               'DefaultConfigVersion': 'Config',
               'DefaultConfigPackage': 'Click',
               'DefaultSoftwareVersion': 'Software',
               'FailOverSE': 'FAIL=SRM',
               'DefaultWhizard2Version': '1.9.5',
               'BasePath': '',
              }
    for opName, value in opsDict.items():
      if args[0].endswith(opName):
        return value
    assert args[0] == opsDict
    return None
  theOps = Mock(name='OpsMock')
  theOps.getValue = mockOpsConfig
  return theOps


@pytest.fixture
def pMockMod():
  """Return Module for ProductionJob."""
  pMockMod = Mock()
  pjMock = Mock(name="ProductionJob")
  pMockMod.return_value = pjMock
  pjMock.getMetadata.return_value = {}
  return pMockMod


@pytest.fixture
def theChain(opsMock):
  """Return production chain fixture."""
  params = Mock()
  params.additionalName = ''
  params.dryRun = True
  with patch("ILCDIRAC.ILCTransformationSystem.scripts.dirac_clic_make_productions.CLICDetProdChain.loadParameters",
             new=Mock()), \
       patch("DIRAC.ConfigurationSystem.Client.Helpers.Operations.Operations",
             new=Mock(return_value=opsMock)):
    chain = theScript.CLICDetProdChain(params)

  return chain


@pytest.fixture
def aTask():
  """Return aTask with minimal content."""
  task = Task({'Energy': 300.0}, {}, 333)
  return task


@pytest.fixture
def cpMock():
  """Return a Mock for the ConfigParser."""
  theCPMock = Mock()
  theCPMock.thisConfigDict = dict(configDict())

  def hasMock(*args, **kwargs):  # pylint: disable=unused-argument
    """Mock the configparser.has_option function."""
    return theCPMock.thisConfigDict.get(args[1])

  def mockConfig(*args, **kwargs):  # pylint: disable=unused-argument
    """Mock the configparser object."""
    assert args[0] == theScript.PP
    return theCPMock.thisConfigDict[args[1]]

  theCPMock.read = Mock()
  theCPMock.get = mockConfig
  theCPMock.has_option = hasMock
  return theCPMock


def test_meta(theChain):
  """Test meta data."""
  ret = theChain.meta(123, 'process', 555.5)
  assert {'ProdID': '123',
          'EvtType': 'process',
          'Energy': '555.5',
          'Machine': 'clic'} == ret


@pytest.mark.parametrize('eInput, eOutput',
                         [(123.0, '123'),
                          (123.03, '123.03'),
                          (123.235, '123.23'),
                          ('123.235', '123.235'),
                          ])
def test_metaEnergy(theChain, eInput, eOutput):
  """Test meta data."""
  assert theChain.metaEnergy(eInput) == eOutput


def test_overlayParameter(theChain):
  """Test overlayPamareters."""
  assert theChain.checkOverlayParameter('300GeV') == '300GeV'
  assert theChain.checkOverlayParameter('3TeV') == '3TeV'
  assert theChain.checkOverlayParameter('') == ''

  with pytest.raises(RuntimeError, match="does not end with unit"):
    theChain.checkOverlayParameter('3000')

  with pytest.raises(RuntimeError, match="does not end with unit"):
    theChain.checkOverlayParameter('3tev')


def test_loadParameters(theChain, cpMock):
  """Test load parameters."""
  parameter = Mock()
  parameter.prodConfigFilename = None
  parameter.dumpConfigFile = None
  theChain.loadParameters(parameter)
  c = theChain

  parameter.prodConfigFilename = 'filename'

  with patch(SCP, new=Mock(return_value=cpMock)):
    c.loadParameters(parameter)
  assert c.prodGroup == "myProdGroup"
  assert c.detectorModel == "myDetectorModel"
  assert c.prodIDs == [123, 456]
  assert c.energies == [100, 200]
  assert c.eventsPerJobs == [1000, 2000]
  assert c.eventsInSplitFiles == [5000, 6000]

  assert c.whizard2Version == "myWhizardVersion"
  assert c.whizard2SinFile == ['myWhizardSinFile1', 'myWhizardSinFile2']

  assert c.moveStatus == 'Active'
  assert c.moveGroupSize == '11'

  cpMock.thisConfigDict['prodIDs'] = "123, 456, 789"
  with patch(SCP, new=Mock(return_value=cpMock)), \
       pytest.raises(AttributeError, match="Lengths of Processes"):
    c.loadParameters(parameter)

  cpMock.thisConfigDict['prodIDs'] = ''
  cpMock.has_option = Mock()
  cpMock.has_option.return_value = False
  with patch(SCP, new=Mock(return_value=cpMock)):
    c.loadParameters(parameter)
  assert c.prodIDs == [1, 1]
  assert c.cliRecoOption == '--Config.Tracking=Tracked'

  cpMock.thisConfigDict['eventsInSplitFiles'] = "1000"
  c._flags._spl = True
  with patch(SCP, new=Mock(return_value=cpMock)), \
       pytest.raises(AttributeError, match="Length of eventsInSplitFiles"):
    c.loadParameters(parameter)

  parameter.prodConfigFilename = None
  parameter.dumpConfigFile = True
  with patch(SCP, new=Mock(return_value=cpMock)), \
    pytest.raises(RuntimeError, match="^$"):
    c.loadParameters(parameter)

  parameter.prodConfigFilename = 'filename'
  parameter.dumpConfigFile = False
  cpMock.thisConfigDict['MoveStatus'] = 'Foo'
  with patch(SCP, new=Mock(return_value=cpMock)), \
       pytest.raises(AttributeError, match='MoveStatus can only be'):
    c.loadParameters(parameter)


def test_createMarlinApplication(theChain, aTask, cpMock):
  """Test creating the marlin application."""
  from ILCDIRAC.Interfaces.API.NewInterface.Applications import Marlin

  parameter = Mock()
  parameter.prodConfigFilename = 'filename'
  parameter.dumpConfigFile = False
  with patch(SCP, new=Mock(return_value=cpMock)):
    theChain.loadParameters(parameter)

  ret = theChain.createMarlinApplication(aTask, over=True)
  assert isinstance(ret, Marlin)
  assert ret.detectortype == 'myDetectorModel'
  assert ret.steeringFile == 'clicReconstruction.xml'
  assert theChain.cliRecoOption == '--Config.Tracking=Tracked'
  assert ret.extraCLIArguments == '--Config.Tracking=Tracked  --Config.Overlay=300GeV'

  with patch(SCP, new=Mock(return_value=cpMock)):
    theChain.loadParameters(parameter)
  theChain._flags._over = False

  ret = theChain.createMarlinApplication(aTask, over=False)
  assert isinstance(ret, Marlin)
  assert ret.detectortype == 'myDetectorModel'
  assert ret.steeringFile == 'clicReconstruction.xml'
  assert theChain.cliRecoOption == '--Config.Tracking=Tracked'
  assert ret.extraCLIArguments == '--Config.Tracking=Tracked'


def test_createWhizard2Application(theChain, aTask, cpMock, opsMock):
  """Test creating the whizard2 application."""
  from ILCDIRAC.Interfaces.API.NewInterface.Applications import Whizard2

  parameter = Mock(name="ParameterMock")
  parameter.whizard2SinFile = 'filename'
  parameter.dumpConfigFile = False
  with patch(SCP, new=Mock(return_value=cpMock)), \
       patch("DIRAC.ConfigurationSystem.Client.Helpers.Operations.Operations",
             new=Mock(return_value=opsMock)):
    theChain.loadParameters(parameter)

  aTask.meta = {'ProdID': '123', 'EvtType': 'process', 'Energy': '555', 'Machine': 'clic'}
  aTask.eventsPerJob = 100
  aTask.genFile = 'sinFile'

  ret = theChain.createWhizard2Application(aTask)
  assert isinstance(ret, Whizard2)
  assert ret.version == 'myWhizardVersion'


def test_createDDSimApplication(theChain, aTask, cpMock, opsMock):
  """Test creating the ddsim application."""
  from ILCDIRAC.Interfaces.API.NewInterface.Applications import DDSim
  parameter = Mock()
  parameter.prodConfigFilename = 'filename'
  parameter.dumpConfigFile = False
  with patch(SCP, new=Mock(return_value=cpMock)), \
       patch("DIRAC.ConfigurationSystem.Client.Helpers.Operations.Operations",
             new=Mock(return_value=opsMock)):
    theChain.loadParameters(parameter)

  ret = theChain.createDDSimApplication(aTask)
  assert isinstance(ret, DDSim)
  assert ret.steeringFile == 'clic_steer.py'


def test_createSplitApplication(theChain, cpMock):
  """Test creating the splitting application."""
  from ILCDIRAC.Interfaces.API.NewInterface.Applications import StdHepSplit

  parameter = Mock()
  parameter.prodConfigFilename = 'filename'
  parameter.dumpConfigFile = False
  with patch(SCP, new=Mock(return_value=cpMock)):
    theChain.loadParameters(parameter)

  ret = theChain.createSplitApplication(100, 1000, 'stdhep')
  assert isinstance(ret, StdHepSplit)
  assert ret.datatype == 'gen'
  assert ret.maxRead == 1000
  assert ret.numberOfEventsPerFile == 100


def test_createOverlayApplication(theChain, aTask, cpMock):
  """Test creating the overlay application."""
  from ILCDIRAC.Interfaces.API.NewInterface.Applications import OverlayInput
  parameter = Mock()
  parameter.prodConfigFilename = 'filename'
  parameter.dumpConfigFile = False
  with patch(SCP, new=Mock(return_value=cpMock)):
    theChain.loadParameters(parameter)
  aTask.meta['Energy'] = 350
  ret = theChain.createOverlayApplication(aTask)
  assert isinstance(ret, OverlayInput)
  assert ret.machine == 'clic_opt'
  aTask.meta['Energy'] = 355
  with pytest.raises(RuntimeError, match='No overlay parameters'):
    ret = theChain.createOverlayApplication(aTask)


def test_createSplitProduction(theChain, pMockMod):
  """Test creating the splitting production."""
  task = Task(metaInput={'ProdID': '23', 'Energy': '350'},
              parameterDict=theChain.getParameterDictionary('MI6')[0],
              eventsPerJob=0o07,
              eventsPerBaseFile=700,
              )
  assert task.meta['NumberOfEvents'] == 700
  with patch("ILCDIRAC.Interfaces.API.NewInterface.ProductionJob.ProductionJob", new=pMockMod):
    retMeta = theChain.createSplitProduction(task)
  assert retMeta == {}


def test_createRecoProduction(theChain, pMockMod):
  """Test creating the reco production."""
  theChain._flags._over = True
  assert theChain._flags.over
  theChain.overlayEvents = '1.4TeV'
  task = Task(metaInput={'ProdID': '23', 'Energy': '350'},
              parameterDict=theChain.getParameterDictionary('MI6')[0],
              eventsPerJob=321,
              )
  with patch("ILCDIRAC.Interfaces.API.NewInterface.ProductionJob.ProductionJob", new=pMockMod):
    retMeta = theChain.createReconstructionProduction(task, over=False)
  assert retMeta == {}
  assert theChain.cliRecoOption == ''


def test_createSimProduction(theChain, pMockMod):
  """Test creating the simulation production."""
  task = Task(metaInput={'ProdID': '23', 'Energy': '350'},
              parameterDict=theChain.getParameterDictionary('MI6')[0],
              eventsPerJob=333,
              )
  with patch("ILCDIRAC.Interfaces.API.NewInterface.ProductionJob.ProductionJob", new=pMockMod):
    retMeta = theChain.createSimulationProduction(task)
  assert retMeta == {}


def test_createGenProduction(theChain, pMockMod):
  """Test creating the generation production."""
  with patch("ILCDIRAC.Interfaces.API.NewInterface.ProductionJob.ProductionJob", new=pMockMod):
    task = Task(metaInput={'ProdID': '23', 'Energy': '350', 'EvtType': 'ttBond'},
                parameterDict=theChain.getParameterDictionary('MI6')[0],
                eventsPerJob=10,
                nbTasks='10',
                genFile='myWhizardSinFile')
    retMeta = theChain.createGenerationProduction(task)
  assert retMeta == {}


def test_createMovingTransformation(theChain):
  """Test creating the moving productions."""
  theChain.outputSE = "Source"
  theChain.finalOutputSE = "Target"
  theChain.moveGroupSize = 13
  theChain._flags._rec = True
  theChain._flags._sim = True
  theChain._flags._moveDst = True
  theChain._flags._moveRec = False
  theChain._flags._moveSim = True
  theChain._flags._moves = True
  theChain._flags._dryRun = False
  with patch("DIRAC.TransformationSystem.Utilities.ReplicationTransformation.createDataTransformation") as moveMock:
    theChain.createMovingTransformation({'ProdID': 666}, 'MCReconstruction')
    parDict = dict(flavour='Moving',
                   targetSE='Target',
                   sourceSE='Source',
                   plugin='Broadcast',
                   metaKey='ProdID',
                   metaValue=666,
                   extraData={'Datatype': 'DST'},
                   tGroup='several',
                   groupSize=13,
                   enable=True,
                   )
    moveMock.assert_called_once_with(**parDict)

  with patch("DIRAC.TransformationSystem.Utilities.ReplicationTransformation.createDataTransformation") as moveMock:
    theChain.createMovingTransformation({'ProdID': 666}, 'MCSimulation')
    parDict = dict(flavour='Moving',
                   targetSE='Target',
                   sourceSE='Source',
                   plugin='BroadcastProcessed',
                   metaKey='ProdID',
                   metaValue=666,
                   extraData={'Datatype': 'SIM'},
                   tGroup='several',
                   groupSize=13,
                   enable=True,
                   )
    moveMock.assert_called_once_with(**parDict)

  theChain._flags._rec = True
  theChain._flags._moves = False
  theChain._flags._dryRun = False
  with patch("DIRAC.TransformationSystem.Utilities.ReplicationTransformation.createDataTransformation") as moveMock:
    theChain.createMovingTransformation({'ProdID': 666}, 'MCReconstruction')
    moveMock.assert_not_called()

  with pytest.raises(RuntimeError, match='ERROR creating Moving'):
    theChain.createMovingTransformation({'ProdID': 666}, "Split")


def test_setApplicationOptions(theChain, aTask):
  """Test setting the application options."""
  application = Mock()
  application.setSomeParameter = Mock()
  aTask.applicationOptions = {'foo': 'bar'}
  theChain.applicationOptions['AppName'] = {'SomeParameter': 'SomeValue', 'FE.foo': ['bar', 'baz'],
                                            'C_Repl': 'longValueWeDoNotwantToRepeat'}
  theChain._setApplicationOptions('AppName', application, aTask.applicationOptions)
  application.setSomeParameter.assert_called_once_with('SomeValue')
  application.setfoo.assert_called_once_with('bar')

  from ILCDIRAC.Interfaces.API.NewInterface.Applications import Marlin
  application = Marlin()
  theChain.applicationOptions['AppName'] = {'SomeOtherParameter': 'SomeValue'}
  with pytest.raises(AttributeError, match='Cannot set'):
    theChain._setApplicationOptions('AppName', application)


def test_getProdInfoFromIDs(theChain):
  """Test getting theproduction information."""
  # successful
  theChain.prodIDs = [12345]
  trClientMock = Mock(name='trClient')
  trClientMock.getTransformation.return_value = S_OK({'EventsPerTask': 123})
  trMock = Mock(return_value=trClientMock)
  fcClientMock = Mock(name='fcClient')
  fcClientMock.findFilesByMetadata.return_value = S_OK(['/path/to/file'])
  fcClientMock.getDirectoryUserMetadata.return_value = S_OK({'EvtType': 'haha', 'Energy': 321})
  fcMock = Mock(return_value=fcClientMock)
  with patch('DIRAC.TransformationSystem.Client.TransformationClient.TransformationClient', new=trMock), \
       patch('DIRAC.Resources.Catalog.FileCatalogClient.FileCatalogClient', new=fcMock):
    theChain._getProdInfoFromIDs()
  assert theChain.eventsPerJobs == [123]
  assert theChain.processes == ['haha']
  assert theChain.energies == [321]

  # first exception
  theChain.prodIDs = []
  with pytest.raises(AttributeError, match='No prodIDs'):
    theChain._getProdInfoFromIDs()

  # second exception
  theChain.prodIDs = [12345]
  trClientMock = Mock(name='trClient')
  trClientMock.getTransformation.return_value = S_ERROR('No such prod')
  trMock = Mock(return_value=trClientMock)
  with patch('DIRAC.TransformationSystem.Client.TransformationClient.TransformationClient', new=trMock), \
       pytest.raises(AttributeError, match='No prodInfo found'):
    theChain._getProdInfoFromIDs()

  # third exception
  theChain.prodIDs = [12345]
  trClientMock = Mock(name='trClient')
  trClientMock.getTransformation.return_value = S_OK({'EventsPerTask': 123})
  trMock = Mock(return_value=trClientMock)
  fcClientMock = Mock(name='fcClient')
  fcClientMock.findFilesByMetadata.return_value = S_ERROR('No files found')
  fcMock = Mock(return_value=fcClientMock)
  with patch('DIRAC.TransformationSystem.Client.TransformationClient.TransformationClient', new=trMock), \
       patch('DIRAC.Resources.Catalog.FileCatalogClient.FileCatalogClient', new=fcMock), \
       pytest.raises(AttributeError, match='Could not find file'):
    theChain._getProdInfoFromIDs()


def test_createTransformations(theChain):
  """Test replcation transformation creation."""
  theChain.createMovingTransformation = Mock(name='MovingTrafo')
  taskDict = dict(MOVE_GEN=[{'move': 'gen'}],
                  MOVE_SPLIT=[{'move': 'split'}],
                  MOVE_SIM=[{'move': 'sim'}],
                  MOVE_REC=[{'move': 'rec'}],
                  MOVE_OVER=[{'move': 'over'}],
                )
  theChain.createTransformations(taskDict)
  theChain.createMovingTransformation.assert_any_call({'move': 'gen'}, 'MCGeneration')
  theChain.createMovingTransformation.assert_any_call({'move': 'split'}, 'MCGeneration')
  theChain.createMovingTransformation.assert_any_call({'move': 'sim'}, 'MCSimulation')
  theChain.createMovingTransformation.assert_any_call({'move': 'rec'}, 'MCReconstruction')
  theChain.createMovingTransformation.assert_any_call({'move': 'over'}, 'MCReconstruction_Overlay')


def test_createTransformations_2(theChain):
  """Test workflow transformation creation."""
  theChain.createMovingTransformation = Mock(name='MovingTrafo')
  theChain.createGenerationProduction = Mock(name='GenTrafo', return_value={'ret': 'gen'})
  theChain.createSimulationProduction = Mock(name='SimTrafo', return_value={'ret': 'sim'})
  theChain.createReconstructionProduction = Mock(name='RecTrafo', return_value={'ret': 'rec'})
  theChain.addSimTask = Mock(name='AddSim')
  theChain.addRecTask = Mock(name='AddRec')

  task = Task(metaInput={'ProdID': '23', 'Energy': '350'},
              parameterDict=theChain.getParameterDictionary('MI6')[0],
              eventsPerJob=0o07,
              eventsPerBaseFile=700,
              )
  taskDict = defaultdict(list)
  taskDict['GEN'].append(task)
  theChain.createTransformations(taskDict)
  theChain.createGenerationProduction.assert_any_call(task)
  theChain.addSimTask.assert_called_with(taskDict, {'ret': 'gen'}, originalTask=task)

  # SIM transformation, off
  theChain.createMovingTransformation.reset_mock()
  theChain.createGenerationProduction.reset_mock()
  theChain.addSimTask.reset_mock()

  taskDict = defaultdict(list)
  taskDict['SIM'].append(task)
  theChain._flags._sim = False
  theChain.createTransformations(taskDict)
  theChain.createSimulationProduction.assert_not_called()
  theChain.addRecTask.assert_not_called()

  # SIM transformation, on
  theChain.createMovingTransformation.reset_mock()
  theChain.createGenerationProduction.reset_mock()
  theChain.addRecTask.reset_mock()

  taskDict = defaultdict(list)
  taskDict['SIM'].append(task)
  theChain._flags._sim = True
  theChain.createTransformations(taskDict)
  theChain.createSimulationProduction.assert_called_with(task)
  theChain.addRecTask.assert_called_with(taskDict, {'ret': 'sim'}, originalTask=task)

  # REC transformation, no over
  taskDict = defaultdict(list)
  taskDict['REC'].append(task)
  theChain._flags._rec = True
  theChain._flags._over = False
  theChain.createTransformations(taskDict)
  theChain.createReconstructionProduction.assert_called_once_with(task, over=False)

  # REC transformation, over
  theChain.createReconstructionProduction.reset_mock()
  taskDict = defaultdict(list)
  taskDict['REC'].append(task)
  theChain._flags._rec = False
  theChain._flags._over = True
  theChain.createTransformations(taskDict)
  theChain.createReconstructionProduction.assert_called_once_with(task, over=True)


def test_addSimTask(theChain):
  """Test adding sim task."""
  taskDict = defaultdict(list)
  theChain.addSimTask(taskDict, metaInput={'ProdID': '23', 'Energy': '350'}, originalTask=Task({}, {}, 123))
  assert len(taskDict['SIM']) == 1

  taskDict = defaultdict(list)
  theChain.applicationOptions['DDSim']['FE.steeringFile'] = ['a.py', 'b.py']
  theChain.applicationOptions['DDSim']['FE.additionalName'] = ['APY', 'BPY']
  theChain.addSimTask(taskDict, metaInput={'ProdID': '23', 'Energy': '350'}, originalTask=Task({}, {}, 123))
  assert len(taskDict['SIM']) == 2
  assert taskDict['SIM'][0].applicationOptions['steeringFile'] == 'a.py'
  assert taskDict['SIM'][1].applicationOptions['steeringFile'] == 'b.py'
  assert taskDict['SIM'][0].taskName == 'APY'
  assert taskDict['SIM'][1].taskName == 'BPY'


def test_addRecTask(theChain):
  """Test adding rec task."""
  taskDict = defaultdict(list)
  theChain.addRecTask(taskDict, metaInput={'ProdID': '23', 'Energy': '350'}, originalTask=Task({}, {}, 123))
  assert len(taskDict['REC']) == 1

  taskDict = defaultdict(list)
  theChain.applicationOptions['Marlin']['FE.steeringFile'] = ['a.xml', 'b.xml']
  theChain.applicationOptions['Marlin']['FE.QueryLanguage'] = ['EN', 'DE']
  theChain.applicationOptions['Marlin']['FE.cliReco'] = ['--Option=Value0', '--Option=Value1']
  theChain.addRecTask(taskDict, metaInput={'ProdID': '23', 'Energy': '350'}, originalTask=Task({}, {}, 123))
  assert len(taskDict['REC']) == 2
  assert taskDict['REC'][0].applicationOptions['steeringFile'] == 'a.xml'
  assert taskDict['REC'][1].applicationOptions['steeringFile'] == 'b.xml'
  assert taskDict['REC'][0].meta['Language'] == 'EN'
  assert taskDict['REC'][1].meta['Language'] == 'DE'
  assert taskDict['REC'][0].cliReco == '--Option=Value0'
  assert taskDict['REC'][1].cliReco == '--Option=Value1'


def test_addGenTask(theChain):
  """Test adding gen task."""
  taskDict = defaultdict(list)
  theChain.addGenTask(taskDict, originalTask=Task({}, {}, 123))
  assert len(taskDict['GEN']) == 1

  taskDict = defaultdict(list)
  theChain.applicationOptions['Whizard2']['FE.additionalName'] = ['1', '2']
  theChain.applicationOptions['Whizard2']['steeringFile'] = ['original.sin']
  theChain.addGenTask(taskDict, originalTask=Task({}, {}, 123))
  assert len(taskDict['GEN']) == 2
  assert taskDict['GEN'][0].taskName == '1'
  assert taskDict['GEN'][1].taskName == '2'


def test_createTaskDict_none(theChain):
  """Test createTaskDict function."""
  taskDict = theChain.createTaskDict(123456, 'ee_qq', 5000, 333, genFile='file.sin', nbTasks=222,
                                     eventsPerBaseFile=None, taskName='')
  for pType in ['GEN', 'SIM', 'REC', 'SPLIT']:
    assert not taskDict[pType]


def test_createTaskDict_gen(theChain):
  """Test createTaskDict function."""
  theChain._flags._gen = True
  taskDict = theChain.createTaskDict(123456, 'ee_qq', 5000, 333, genFile='file.sin', nbTasks=222,
                                     eventsPerBaseFile=None, taskName='')
  assert len(taskDict['GEN']) == 1
  assert taskDict['GEN'][0].genFile == 'file.sin'
  assert taskDict['GEN'][0].nbTasks == 222
  assert taskDict['GEN'][0].meta == {'EvtType': 'ee_qq', 'ProdID': '123456',
                                     'Machine': 'clic',
                                     'NumberOfEvents': 333,
                                     'Energy': '5000',
                                     }


def test_createTaskDict_sim(theChain):
  """Test createTaskDict function."""
  theChain._flags._sim = True
  taskDict = theChain.createTaskDict(123456, 'ee_qqqq', 5000, 333, genFile=None,
                                     nbTasks=None, eventsPerBaseFile=None, taskName='')
  assert len(taskDict['SIM']) == 1
  assert taskDict['SIM'][0].genFile is None
  assert taskDict['SIM'][0].nbTasks is None
  assert taskDict['SIM'][0].eventsPerJob == 333


def test_createTaskDict_sim_split(theChain):
  """Test createTaskDict function."""
  theChain._flags._sim = True
  theChain._flags._spl = True
  # no need to split
  taskDict = theChain.createTaskDict(123456, 'ee_qqqq', 5000, 25, genFile=None, nbTasks=None,
                                     eventsPerBaseFile=25, taskName='')
  assert not taskDict['SPLIT']
  assert len(taskDict['SIM']) == 1
  # need to split
  taskDict = theChain.createTaskDict(123456, 'ee_qqqq', 5000, 25, genFile=None, nbTasks=None,
                                     eventsPerBaseFile=400, taskName='')
  assert len(taskDict['SPLIT']) == 1
  assert not taskDict['SIM']


def test_createTaskDict_nosplit(theChain):
  """Test createTaskDict function."""
  theChain._flags._spl = True
  taskDict = theChain.createTaskDict(123456, 'ee_qqqq', 5000, 25, genFile=None, nbTasks=None,
                                     eventsPerBaseFile=25, taskName='')
  assert not taskDict['SPLIT']
  assert not taskDict['SIM']


def test_createTaskDict_rec(theChain):
  """Test createTaskDict function."""
  theChain._flags._rec = True
  theChain.applicationOptions['Marlin']['FE.QueryMachine'] = 'fccee, clic'
  taskDict = theChain.createTaskDict(123456, 'ee_qqqq', 5000, 25, genFile=None, nbTasks=None,
                                     eventsPerBaseFile=400, taskName='')
  assert len(taskDict['REC']) == 2
  assert taskDict['REC'][0].meta['Machine'] == 'fccee'
  assert taskDict['REC'][1].meta['Machine'] == 'clic'


@pytest.fixture
def theFlags():
  """Return the flags fixture."""
  return theScript.CLICDetProdChain.Flags()


def test_flags_init(theFlags):
  """Returns the flags constructor."""
  f = theFlags
  assert f._dryRun
  assert not f._gen
  assert not f._spl
  assert not f._sim
  assert not f._rec
  assert not f._over
  assert not f._moves
  assert not f._moveGen
  assert not f._moveSim
  assert not f._moveRec
  assert not f._moveDst


def test_flags_properties(theFlags):
  """Test the flags properties."""
  f = theFlags
  f._gen = True
  f._spl = True
  f._sim = True
  f._rec = False
  f._over = True
  assert f.dryRun
  assert f.gen
  assert f.spl
  assert f.sim
  assert not f.rec
  assert f.over
  f._dryRun = True
  f._moves = True
  f._moveRec = True
  assert f.move
  assert not f.moveGen
  assert not f.moveSim
  assert f.moveRec
  assert not f.moveDst
  f._dryRun = False
  f._moveGen = True
  f._moveSim = True
  f._moveRec = True
  f._moveDst = False
  assert f.move
  assert f.moveGen
  assert f.moveSim
  assert f.moveRec
  assert not f.moveDst


def test_flags_str(theFlags):
  """Test the flags string representation."""
  theFlags._gen = True
  theFlags._sim = True
  theFlags._rec = False
  theFlags._over = True
  flagStr = str(theFlags)
  assert flagStr == """

#Productions to create: Gen, Split, Sim, Rec, RecOver
ProdTypes = Gen, Sim, RecOver

move = False

#Datatypes to move: Gen, Sim, Rec, Dst
MoveTypes = \n"""


def test_loadFlags(theFlags):
  """Test loading the flags from config."""
  myConfig = six.moves.configparser.ConfigParser()
  myConfig.add_section(theScript.PRODUCTION_PARAMETERS)
  myConfig.set(theScript.PRODUCTION_PARAMETERS, 'ProdTypes', 'Gen, Sim,Rec')
  myConfig.set(theScript.PRODUCTION_PARAMETERS, 'move', 'False')
  myConfig.set(theScript.PRODUCTION_PARAMETERS, 'MoveTypes', 'gen, dst')
  theFlags.loadFlags(myConfig)
  f = theFlags
  assert f.gen
  assert f.sim
  assert f.rec
  assert not f.over
  assert f._moveGen
  assert not f._moveSim
  assert not f._moveRec
  assert f._moveDst
  myConfig.set(theScript.PRODUCTION_PARAMETERS, 'MoveTypes', 'gen, dst, badType')
  with pytest.raises(AttributeError, match='badType'):
    theFlags.loadFlags(myConfig)


@pytest.fixture
def theParams():
  """Return the Params fixture."""
  return theScript._Params()


def test_params_init(theParams):
  """Test the Params constructor."""
  assert theParams.prodConfigFilename is None
  assert not theParams.dumpConfigFile
  assert theParams.dryRun
  assert theParams.additionalName == ''


def test_params_settters(theParams):
  """Test the Params setters."""
  with patch("%s.os.path.exists" % THE_SCRIPT, new=Mock(return_value=True)):
    assert theParams.setProdConf('myconf')['OK']

  assert theParams.prodConfigFilename, 'myconf'
  assert theParams.setDumpConf('_')['OK']
  assert theParams.dumpConfigFile
  assert theParams.setEnable('_')['OK']
  assert not theParams.dryRun
  assert theParams.setAddName('addName')['OK']
  assert theParams.additionalName, 'addName'



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""tests for OverlayFiles module."""

from __future__ import absolute_import
import unittest


from ILCDIRAC.Core.Utilities import OverlayFiles as module


class TestHelper(unittest.TestCase):
  """Test helper functions in the script."""

  def setUp(self):
    pass

  def test_energyWithUnit(self):
    self.assertEqual(module.energyWithUnit(300.0), '300GeV')
    self.assertEqual(module.energyWithUnit(380.0), '380GeV')
    self.assertEqual(module.energyWithUnit(3000.0), '3TeV')
    self.assertEqual(module.energyWithUnit(1000.0), '1TeV')
    self.assertEqual(module.energyWithUnit(1400.0), '1.4TeV')
    self.assertEqual(module.energyWithUnit(2500.0), '2.5TeV')

  def test_backwardcompatibility(self):
    self.assertEqual(module.energyWithLowerCaseUnit(300.0), module.oldEnergyWithUnit(300.0))
    self.assertEqual(module.energyWithLowerCaseUnit(380.0), module.oldEnergyWithUnit(380.0))
    self.assertEqual(module.energyWithLowerCaseUnit(3000.0), module.oldEnergyWithUnit(3000.0))
    self.assertEqual(module.energyWithLowerCaseUnit(1000.0), module.oldEnergyWithUnit(1000.0))
    self.assertEqual(module.energyWithLowerCaseUnit(1400.0), module.oldEnergyWithUnit(1400.0))
    self.assertEqual(module.energyWithLowerCaseUnit(2500.0), module.oldEnergyWithUnit(2500.0))

  def test_energyToInt(self):
    self.assertEqual(module.energyToInt('300GeV'), 300)
    self.assertEqual(module.energyToInt('380GeV'), 380)
    self.assertEqual(module.energyToInt('3TeV'), 3000)
    self.assertEqual(module.energyToInt('1TeV'), 1000)
    self.assertEqual(module.energyToInt('1.4TeV'), 1400)
    self.assertEqual(module.energyToInt('2.5TeV'), 2500)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

########################################################################
# File :    DownloadInputData.py
# Author :  Stuart Paterson
########################################################################

"""The Download Input Data module wraps around the Replica Management components to provide access to datasets by available site protocols as defined in the CS for the VO."""

from __future__ import absolute_import
__RCSID__ = "$Id$"

import os
import tempfile
import random

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.WorkloadManagementSystem.Client.JobStateUpdateClient import JobStateUpdateClient
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.Core.Utilities.Os import getDiskSpace
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers

COMPONENT_NAME = 'DownloadInputData'


def _isCached(lfn, seName):
  result = returnSingleResult(StorageElement(seName).getFileMetadata(lfn))
  if not result['OK']:
    return False
  metadata = result['Value']
  return metadata.get('Cached', metadata['Accessible'])


class DownloadInputData(object):
  """retrieve InputData LFN from localSEs (if available) or from elsewhere."""

  #############################################################################
  def __init__(self, argumentsDict):
    """Standard constructor."""
    self.name = COMPONENT_NAME
    self.log = gLogger.getSubLogger(self.name)
    self.inputData = argumentsDict['InputData']
    self.configuration = argumentsDict['Configuration']
    # Warning: this contains not only the SEs but also the file metadata
    self.fileCatalogResult = argumentsDict['FileCatalog']
    # By default put each input data file into a separate directory
    self.inputDataDirectory = argumentsDict.get('InputDataDirectory', 'PerFile')
    self.jobID = None
    self.counter = 1
    self.availableSEs = DMSHelpers().getStorageElements()

  #############################################################################
  def execute(self, dataToResolve=None):
    """This method is called to download the requested files in the case where enough local disk space is available.

    A buffer is left in this calculation
    to leave room for any produced files.
    """

    # Define local configuration options present at every site
    localSESet = set(self.configuration['LocalSEList'])

    self.jobID = self.configuration.get('JobID')

    if dataToResolve:
      self.log.verbose('Data to resolve passed directly to DownloadInputData module')
      self.inputData = dataToResolve  # e.g. list supplied by another module

    self.inputData = sorted(lfn.replace('LFN:', '') for lfn in self.inputData)
    self.log.info('InputData to be downloaded is:\n%s' % '\n'.join(self.inputData))

    replicas = self.fileCatalogResult['Value']['Successful']

    # Problematic files will be returned and can be handled by another module
    failedReplicas = set()
    # For the case that a file is found on two SEs at the same site
    # disk-based replicas are favoured.
    downloadReplicas = {}

    for lfn, reps in replicas.items():
      if lfn not in self.inputData:
        self.log.verbose('LFN %s is not in requested input data to download')
        failedReplicas.add(lfn)
        continue

      if not ('Size' in reps and 'GUID' in reps):
        self.log.error('Missing LFN metadata', "%s %s" % (lfn, str(reps)))
        failedReplicas.add(lfn)
        continue

      # Get and remove size and GUID
      size = reps.pop('Size')
      guid = reps.pop('GUID')
      # Remove all other items that are not SEs
      for item in list(reps.keys()):
        if item not in self.availableSEs:
          reps.pop(item)
      downloadReplicas[lfn] = {'SE': [], 'Size': size, 'GUID': guid}
      # First get Disk replicas
      localReps = set(reps) & localSESet
      for seName in localReps:
        seStatus = StorageElement(seName).status()
        if seStatus['DiskSE'] and seStatus['Read']:
          downloadReplicas[lfn]['SE'].append(seName)
      # If no disk replicas, take tape replicas
      if not downloadReplicas[lfn]['SE']:
        for seName in localReps:
          seStatus = StorageElement(seName).status()
          if seStatus['TapeSE'] and seStatus['Read'] and _isCached(lfn, seName):
            # Only consider replicas that are cached
            downloadReplicas[lfn]['SE'].append(seName)

    totalSize = 0
    verbose = self.log.verbose('Replicas to download are:')
    for lfn, reps in downloadReplicas.items():
      self.log.verbose(lfn)
      if not reps['SE']:
        self.log.info('Failed to find data at local SEs, will try to download from anywhere', lfn)
        reps['SE'] = ''
      else:
        if len(reps['SE']) > 1:
          # if more than one SE is available randomly select one
          random.shuffle(reps['SE'])
        reps['SE'] = reps['SE'][0]
      totalSize += int(reps.get('Size', 0))
      if verbose:
        for item, value in sorted(reps.items()):
          if value:
            self.log.verbose('\t%s %s' % (item, value))

    self.log.info("Total size of files to be downloaded", "is %s bytes" % totalSize)
    for lfn in failedReplicas:
      self.log.warn('Not all file metadata (SE,PFN,Size,GUID) was available for LFN', lfn)

    # Now need to check that the list of replicas to download fits into
    # the available disk space. Initially this is a simple check and if there is not
    # space for all input data, no downloads are attempted.
    result = self.__checkDiskSpace(totalSize)
    if not result['OK']:
      self.log.warn('Problem checking available disk space:\n%s' % (result))
      return result

    # FIXME: this can never happen at the moment
    if not result['Value']:
      self.log.warn("Not enough disk space available for download",
                    "%s / %s bytes" % (result['Value'], totalSize))
      self.__setJobParam(COMPONENT_NAME,
                         "Not enough disk space available for download: %s / %s bytes" % (result['Value'],
                                                                                          totalSize))
      return S_OK({'Failed': self.inputData, 'Successful': {}})

    resolvedData = {}
    localSECount = 0
    for lfn, info in downloadReplicas.items():
      seName = info['SE']
      guid = info['GUID']
      reps = replicas.get(lfn, {})
      if seName:
        result = returnSingleResult(StorageElement(seName).getFileMetadata(lfn))
        if not result['OK']:
          self.log.error("Error getting metadata", result['Message'])
          error = result['Message']
        else:
          metadata = result['Value']
          if metadata.get('Lost', False):
            error = "PFN has been Lost by the StorageElement"
          elif metadata.get('Unavailable', False):
            error = "PFN is declared Unavailable by the StorageElement"
          elif not metadata.get('Cached', metadata['Accessible']):
            error = "PFN is no longer in StorageElement Cache"
          else:
            error = ''
        if error:
          self.log.error(error, lfn)
          result = {'OK': False}
        else:
          self.log.info('Preliminary checks OK', 'download %s from %s:' % (lfn, seName))
          result = self._downloadFromSE(lfn, seName, reps, guid)
          if not result['OK']:
            self.log.error("Download failed", "Tried downloading from SE %s: %s" % (seName, result['Message']))
      else:
        result = {'OK': False}

      if not result['OK']:
        reps.pop(seName, None)
        # Check the other SEs
        if reps:
          self.log.info('Trying to download from any SE')
          result = self._downloadFromBestSE(lfn, reps, guid)
          if not result['OK']:
            self.log.error("Download from best SE failed", "Tried downloading %s: %s" % (lfn, result['Message']))
            failedReplicas.add(lfn)
        else:
          failedReplicas.add(lfn)
      else:
        localSECount += 1
      if result['OK']:
        # Rename file if downloaded FileName does not match the LFN... How can this happen?
        lfnName = os.path.basename(lfn)
        oldPath = result['Value']['path']
        fileName = os.path.basename(oldPath)
        if lfnName != fileName:
          newPath = os.path.join(os.path.dirname(oldPath), lfnName)
          os.rename(oldPath, newPath)
          result['Value']['path'] = newPath
        resolvedData[lfn] = result['Value']

    # Report datasets that could not be downloaded
    report = ''
    if resolvedData:
      report += 'Successfully downloaded %d LFN(s)' % len(resolvedData)
      if localSECount != len(resolvedData):
        report += ' (%d from local SEs):\n' % localSECount
      else:
        report += ' from local SEs:\n'
      report += '\n'.join(sorted(resolvedData))
    failedReplicas = sorted(failedReplicas.difference(resolvedData))
    if failedReplicas:
      self.log.warn('The following LFN(s) could not be downloaded to the WN:\n%s' % 'n'.join(failedReplicas))
      report += '\nFailed to download %d LFN(s):\n' % len(failedReplicas)
      report += '\n'.join(failedReplicas)

    if report:
      self.__setJobParam(COMPONENT_NAME, report)

    return S_OK({'Successful': resolvedData, 'Failed': failedReplicas})

  #############################################################################
  def __checkDiskSpace(self, totalSize):
    """Compare available disk space to the file size reported from the catalog result."""
    diskSpace = getDiskSpace(self.__getDownloadDir(False))  # MB
    availableBytes = diskSpace * 1024 * 1024  # bytes
    # below can be a configuration option sent via the job wrapper in the future
    # Moved from 3 to 5 GB (PhC 130822) for standard output file
    bufferGBs = 5.0
    data = bufferGBs * 1024 * 1024 * 1024  # bufferGBs in bytes
    if (data + totalSize) < availableBytes:
      msg = 'Enough disk space available (%s bytes)' % (availableBytes)
      self.log.verbose(msg)
      return S_OK(msg)
    else:
      msg = 'Not enough disk space available for download %s (including %dGB buffer) > %s bytes' \
          % ((data + totalSize), bufferGBs, availableBytes)
      self.log.warn(msg)
      return S_ERROR(msg)

  def __getDownloadDir(self, incrementCounter=True):
    if self.inputDataDirectory == "PerFile":
      if incrementCounter:
        self.counter += 1
      return tempfile.mkdtemp(prefix='InputData_%s' % (self.counter), dir=os.getcwd())
    elif self.inputDataDirectory == "CWD":
      return os.getcwd()
    else:
      return self.inputDataDirectory

  #############################################################################
  def _downloadFromBestSE(self, lfn, reps, guid):
    """Download a local copy of a single LFN from a list of Storage Elements.

    This is used as a last resort to attempt to retrieve the file.
    """
    self.log.verbose("Attempting to download file from all SEs", "(%s): %s" % (','.join(reps), lfn))
    diskSEs = set()
    tapeSEs = set()
    # Sort replicas, disk first
    for seName in reps:
      seStatus = StorageElement(seName).status()
      # FIXME: This is simply terrible - this notion of "DiskSE" vs "TapeSE" should NOT be used here!
      if seStatus['Read'] and seStatus['DiskSE']:
        diskSEs.add(seName)
      elif seStatus['Read'] and seStatus['TapeSE']:
        tapeSEs.add(seName)

    for seName in list(diskSEs) + list(tapeSEs):
      if seName in diskSEs or _isCached(lfn, seName):
        # On disk or cached from tape
        result = self._downloadFromSE(lfn, seName, reps, guid)
        if result['OK']:
          return result
        else:
          self.log.error("Download failed", "Tried downloading %s from SE %s: %s" % (lfn, seName, result['Message']))

    return S_ERROR("Unable to download the file from any SE")

  #############################################################################
  def _downloadFromSE(self, lfn, seName, reps, guid):
    """Download a local copy from the specified Storage Element."""
    if not lfn:
      return S_ERROR("LFN not specified: assume file is not at this site")

    self.log.verbose("Attempting to download file", "%s from %s:" % (lfn, seName))

    downloadDir = self.__getDownloadDir()
    fileName = os.path.basename(lfn)
    for localFile in (os.path.join(os.getcwd(), fileName), os.path.join(downloadDir, fileName)):
      if os.path.exists(localFile):
        self.log.info("File already exists locally", "%s as %s" % (fileName, localFile))
        fileDict = {'turl': 'LocalData',
                    'protocol': 'LocalData',
                    'se': seName,
                    'pfn': reps[seName],
                    'guid': guid,
                    'path': localFile}
        return S_OK(fileDict)

    localFile = os.path.join(downloadDir, fileName)
    result = returnSingleResult(StorageElement(seName).getFile(lfn, localPath=downloadDir))
    if not result['OK']:
      self.log.warn('Problem getting lfn', '%s from %s:\n%s' % (lfn, seName, result['Message']))
      self.__cleanFailedFile(lfn, downloadDir)
      return result

    if os.path.exists(localFile):
      self.log.verbose("File successfully downloaded locally", "(%s to %s)" % (lfn, localFile))
      fileDict = {'turl': 'Downloaded',
                  'protocol': 'Downloaded',
                  'se': seName,
                  'pfn': reps[seName],
                  'guid': guid,
                  'path': localFile}
      return S_OK(fileDict)
    else:
      self.log.warn('File does not exist in local directory after download')
      return S_ERROR('OK download result but file missing in current directory')

  #############################################################################
  def __setJobParam(self, name, value):
    """Wraps around setJobParameter of state update client."""
    if not self.jobID:
      return S_ERROR('JobID not defined')

    self.log.verbose('setting job parameters', 'setJobParameter(%s,%s,%s)' % (self.jobID, name, value))
    jobParam = JobStateUpdateClient().setJobParameter(int(self.jobID), str(name), str(value))
    if not jobParam['OK']:
      self.log.warn("Failed to set job parameters", jobParam['Message'])

    return jobParam

  def __cleanFailedFile(self, lfn, downloadDir):
    """Try to remove a file after a failed download attempt."""
    filePath = os.path.join(downloadDir, os.path.basename(lfn))
    self.log.error("Trying to remove file after failed download", "Local path: %s " % filePath)
    if os.path.exists(filePath):
      try:
        os.remove(filePath)
        self.log.error("Removed file remnant after failed download", "Local path: %s " % filePath)
      except OSError as e:
        self.log.error("Failed to remove file after failed download", repr(e))

# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Find files in the dirac file catalog based on meta data.

Usage::

   dirac-ilc-find-in-FC [-D] PATH Constraint1 [Constraint2 [...]]

It is also possible to use any of these operators >=, <=, >, <, !=, = when using metadata constraints.  The list of
metadata options can be obtained from the :doc:`UserGuide/CommandReference/DataManagement/dirac-dms-filecatalog-cli`
by typing: ``meta show``

For example::

   dirac-ilc-find-in-FC /ilc ProdID>1234 Datatype=DST

to list only the directories containing the files use the "-D" flag::

   dirac-ilc-find-in-FC -D /ilc ProdID>1234 Datatype=DST


.. seealso::

  :doc:`UserGuide/CommandReference/DataManagement/dirac-dms-find-lfns`


:since: Mar 20, 2013
:author: stephane
"""

from __future__ import print_function
from __future__ import absolute_import
from DIRAC.Core.Base import Script
from DIRAC.Core.Utilities.DIRACScript import DIRACScript
from DIRAC import gLogger, S_OK
from DIRAC.Core.Utilities.List import uniqueElements
from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient

LOG = gLogger.getSubLogger('')

__RCSID__ = "$Id$"

OPLIST = ['>=', '<=', '>', '<', '!=', '=']
SCRIPTNAME = "dirac-ilc-find-in-FC"


class _Params(object):
  """Parameter Object."""

  def __init__(self):
    self.printOnlyDirectories = False

  def setPrintOnlyDs(self, dummy_opt):
    self.printOnlyDirectories = True
    return S_OK()

  def registerSwitches(self):
    Script.registerSwitch("D", "OnlyDirectories", "Print only directories", self.setPrintOnlyDs)
    Script.setUsageMessage(
        """%s [-D] path meta1=A meta2=B etc.\nPossible operators for metadata: %s""" %
        (SCRIPTNAME, OPLIST))


def _createQueryDict(argss):
  """Create a proper dictionary, stolen from FC CLI."""

  fileCatClient = FileCatalogClient()
  result = fileCatClient.getMetadataFields()

  if not result['OK']:
    LOG.error("Failed checking for metadata fields")
    return None
  if not result['Value']:
    LOG.error('No meta data fields available')
    return None
  typeDict = result['Value']['FileMetaFields']
  typeDict.update(result['Value']['DirectoryMetaFields'])
  metaDict = {}
  contMode = False
  for arg in argss:
    if not contMode:
      operation = ''
      for op in OPLIST:
        if arg.find(op) != -1:
          operation = op
          break
      if not operation:
        LOG.error("Error: operation is not found in the query")
        return None

      name, value = arg.split(operation)
      if name not in typeDict:
        LOG.error("Error: metadata field %s not defined" % name)
        return None
      mtype = typeDict[name]
    else:
      value += ' ' + arg
      value = value.replace(contMode, '')
      contMode = False

    if value[0] in ['"', "'"]:
      if value[-1] not in ['"', "'"]:
        contMode = value[0]
        continue

    if value.find(',') != -1:
      valueList = [x.replace("'", "").replace('"', '') for x in value.split(',')]
      mvalue = valueList
      if mtype[0:3].lower() == 'int':
        mvalue = [int(x) for x in valueList if x not in ['Missing', 'Any']]
        mvalue += [x for x in valueList if x in ['Missing', 'Any']]
      if mtype[0:5].lower() == 'float':
        mvalue = [float(x) for x in valueList if x not in ['Missing', 'Any']]
        mvalue += [x for x in valueList if x in ['Missing', 'Any']]
      if operation == "=":
        operation = 'in'
      if operation == "!=":
        operation = 'nin'
      mvalue = {operation: mvalue}
    else:
      mvalue = value.replace("'", "").replace('"', '')
      if value not in ['Missing', 'Any']:
        if mtype[0:3].lower() == 'int':
          mvalue = int(value)
        if mtype[0:5].lower() == 'float':
          mvalue = float(value)
      if operation != '=':
        mvalue = {operation: mvalue}

    if name in metaDict:
      if isinstance(metaDict[name], dict):
        if isinstance(mvalue, dict):
          op, value = list(mvalue.items())[0]
          if op in metaDict[name]:
            if isinstance(metaDict[name][op], list):
              if isinstance(value, list):
                metaDict[name][op] = uniqueElements(metaDict[name][op] + value)
              else:
                metaDict[name][op] = uniqueElements(metaDict[name][op].append(value))
            else:
              if isinstance(value, list):
                metaDict[name][op] = uniqueElements([metaDict[name][op]] + value)
              else:
                metaDict[name][op] = uniqueElements([metaDict[name][op], value])
          else:
            metaDict[name].update(mvalue)
        else:
          if isinstance(mvalue, list):
            metaDict[name].update({'in': mvalue})
          else:
            metaDict[name].update({'=': mvalue})
      elif isinstance(metaDict[name], list):
        if isinstance(mvalue, dict):
          metaDict[name] = {'in': metaDict[name]}
          metaDict[name].update(mvalue)
        elif isinstance(mvalue, list):
          metaDict[name] = uniqueElements(metaDict[name] + mvalue)
        else:
          metaDict[name] = uniqueElements(metaDict[name].append(mvalue))
      else:
        if isinstance(mvalue, dict):
          metaDict[name] = {'=': metaDict[name]}
          metaDict[name].update(mvalue)
        elif isinstance(mvalue, list):
          metaDict[name] = uniqueElements([metaDict[name]] + mvalue)
        else:
          metaDict[name] = uniqueElements([metaDict[name], mvalue])
    else:
      metaDict[name] = mvalue

  return metaDict


@DIRACScript()
def main():
  """Find something in the FileCatalog."""
  from DIRAC import exit as dexit
  clip = _Params()
  clip.registerSwitches()
  Script.parseCommandLine()

  args = Script.getPositionalArgs()
  if len(args) < 2:
    Script.showHelp('ERROR: Not enough arguments')
    LOG.error("Run %s --help" % SCRIPTNAME)
    dexit(1)

  path = args[0]
  if path == '.':
    path = '/'

  # Check that the first argument is not a MetaQuery
  if any(op in path for op in OPLIST):
    LOG.error("ERROR: Path '%s' is not a valid path! The first argument must be a path" % path)
    LOG.error("Run %s --help" % SCRIPTNAME)
    dexit(1)

  LOG.verbose("Path:", path)
  metaQuery = args[1:]
  metaDataDict = _createQueryDict(metaQuery)
  LOG.verbose("Query:", str(metaDataDict))
  if not metaDataDict:
    LOG.info("No query")
    dexit(1)

  fc = FileCatalogClient()
  res = fc.findFilesByMetadata(metaDataDict, path)
  if not res['OK']:
    LOG.error(res['Message'])
    dexit(1)
  if not res['Value']:
    LOG.notice("No files found")

  listToPrint = None

  if clip.printOnlyDirectories:
    listToPrint = set("/".join(fullpath.split("/")[:-1]) for fullpath in res['Value'])
  else:
    listToPrint = res['Value']

  for entry in listToPrint:
    print(entry)

  dexit(0)


if __name__ == '__main__':
  main()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Run Pandora Calorimeter calibration.

Runs on each worker node.
ILCDIRAC.Workflow.Modules.Calibration called by Job Agent.

:since: March, 2019

:author: Jan Hendrik Ebbing
:author: Oleksandr Viazlo
"""

from __future__ import absolute_import
import os
import shutil

from ILCDIRAC.Core.Utilities.WasteCPU import wasteCPUCycles
from DIRAC.Core.Utilities.Subprocess import shellCall
from DIRAC import S_OK, S_ERROR, gLogger

from ILCDIRAC.Core.Utilities.CombinedSoftwareInstallation import getEnvironmentScript
from ILCDIRAC.Core.Utilities.resolvePathsAndNames import resolveIFpaths
from ILCDIRAC.Core.Utilities.FindSteeringFileDir import getSteeringFileDirName
from ILCDIRAC.Workflow.Modules.MarlinAnalysis import MarlinAnalysis
from ILCDIRAC.CalibrationSystem.Client.CalibrationClient import CalibrationClient
from ILCDIRAC.CalibrationSystem.Service.CalibrationRun import CalibrationPhase
from ILCDIRAC.CalibrationSystem.Utilities.functions import updateSteeringFile
from ILCDIRAC.CalibrationSystem.Utilities.functions import readValueFromSteeringFile
from ILCDIRAC.CalibrationSystem.Utilities.functions import addPfoAnalysisProcessor
from ILCDIRAC.CalibrationSystem.Utilities.fileutils import stringToBinaryFile
from ILCDIRAC.CalibrationSystem.Utilities.functions import readParametersFromSteeringFile
from ILCDIRAC.CalibrationSystem.Utilities.functions import addParameterToProcessor

__RCSID__ = '$Id$'


class Calibration(MarlinAnalysis):
  """Define the Calibration part of the workflow."""

  def __init__(self):
    """Initialize."""
    super(Calibration, self).__init__()
    self.applicationName = "Calibration"
    self.currentStep = None
    self.currentPhase = None
    self.currentStage = None
    self.calibrationID = None
    self.workerID = None
    self.cali = None
    self.log = None

  def runIt(self):
    """Execute.

    Called by Job Agent

    Execute the following:
      1) resolve input data and environment
      2) request parameters from the service
      3) prepare Marlin steering file
      4) run Marlin and catch the exit code
      5) report results back to service
      6) repeat steps 2-5 until calibration is finished

    :return: S_OK(), S_ERROR()
    """
    self.setApplicationStatus('PandoraCalib_%s: setting up jobs' % self.calibrationID)

    if self.cali is None:
      self.cali = CalibrationClient(self.calibrationID, self.workerID)
    if self.log is None:
      self.log = gLogger.getSubLogger('%s_wid_%s' % (__name__, self.workerID))

    self.result = S_OK()
    if not self.platform:
      self.result = S_ERROR('No ILC platform selected')
    elif not self.applicationLog:
      self.result = S_ERROR('No Log file provided')
    if not self.result['OK']:
      self.log.error("Failed to resolve input parameters:", self.result["Message"])
      return self.result

    if not self.workflowStatus['OK'] or not self.stepStatus['OK']:
      self.log.verbose('Workflow status = %s, step status = %s' % (self.workflowStatus['OK'], self.stepStatus['OK']))
      return S_OK('%s should not proceed as previous step did not end properly' % self.applicationName)

    # FIXME this file is needed to disable watchdog check to prevent killing jobs
    self.log.info('creating DISABLE_WATCHDOG_CPU_WALLCLOCK_CHECK file to disable watchdog')
    fopen = open('DISABLE_WATCHDOG_CPU_WALLCLOCK_CHECK', 'w')
    fopen.close()

    # get the path to the detector model, either local or from the software
    compactFile = None
    if self.detectorModel:
      resXML = self._getDetectorXML()
      if not resXML['OK']:
        self.log.error("Could not obtain the detector XML file: ", resXML["Message"])
        return resXML
      compactFile = resXML['Value']
    else:
      self.log.error('no detectorModel specified! use model which is provided in the template steering file')

    res = getEnvironmentScript(self.platform, "marlin", self.applicationVersion, self.getEnvScript)
    if not res['OK']:
      self.log.error("Failed to get the env script")
      return res
    env_script_path = res["Value"]

    res = resolveIFpaths(self.InputData)
    if not res['OK']:
      self.log.error("Failed to resolve path to input slcio files: %s" % res)
      return res
    listofslcio = res['Value']
    self.log.info('SASHA resolved input data: %s' % listofslcio)

    steeringfiledirname = ''
    res = getSteeringFileDirName(self.platform, "marlin", self.applicationVersion)
    if res['OK']:
      steeringfiledirname = res['Value']
    else:
      self.log.warn('Could not find the steering file directory', res['Message'])

    # Handle PandoraSettings.xml
    # TODO directory below is detector dependent... implement it
    pandorasettings = 'PandoraSettings/PandoraSettings.xml'
    if 'FCC' in self.detectorModel:  # pylint: disable=unsupported-membership-test
      pandorasettings = 'PandoraSettingsFCCee/PandoraSettings.xml'
    if not os.path.exists(pandorasettings):
      if steeringfiledirname and os.path.exists(os.path.join(steeringfiledirname, pandorasettings)):
        try:
          shutil.copy(os.path.join(steeringfiledirname, pandorasettings),
                      os.path.join(os.getcwd(), pandorasettings))
        except EnvironmentError as x:
          self.log.warn('Could not copy PandoraSettings.xml, exception: %s' % x)

    # Handle PandoraSettingsPhotonTraining.xml which is used for photon training stage
    pandorasettings = 'CalibrationPandoraSettings/PandoraSettingsPhotonTraining.xml'
    if not os.path.exists(pandorasettings):
      # FIXME is this path wrong?
      photontrainingfiledirname = os.path.join(steeringfiledirname, '../CalibrationPandoraSettings/')
      if photontrainingfiledirname and os.path.exists(os.path.join(photontrainingfiledirname, pandorasettings)):
        try:
          fullPathPandoraSettings = os.path.join(os.getcwd(), pandorasettings)
          shutil.copy(os.path.join(photontrainingfiledirname, pandorasettings),
                      fullPathPandoraSettings)
        except EnvironmentError as x:
          self.log.warn('Could not copy and prepare PandoraSettingsPhotonTraining.xml, exception: %s' % x)
    # rename output xml-file from photon training stage
    updateSteeringFile(pandorasettings, pandorasettings,
                       {"algorithm[@type='PhotonReconstruction']/HistogramFile":
                        'PandoraLikelihoodDataPhotonTraining.xml'})

    if not os.path.exists(self.SteeringFile):
      if steeringfiledirname:
        if os.path.exists(os.path.join(steeringfiledirname, self.SteeringFile)):
          self.SteeringFile = os.path.join(steeringfiledirname, self.SteeringFile)
          # default steering file doesn't have PfoAnalysis processor
    if not self.SteeringFile:
      self.log.error("Steering file not defined, this shouldn't happen!")
      return S_ERROR("Could not find steering file")
    if not os.path.exists(self.SteeringFile):
      self.log.error("Steering is not found!")
      return S_ERROR("Could not find steering file")
    self.log.info("Steering file: %s" % self.SteeringFile)

    # check if steering file contains PfoAnalysis processor needed for calibration
    if readValueFromSteeringFile(self.SteeringFile, ".//processor[@type='PfoAnalysis']") is None:
      addPfoAnalysisProcessor(self.SteeringFile)

    res = self.prepareMARLIN_DLL(env_script_path)
    if not res['OK']:
      self.log.error('Failed building MARLIN_DLL:', res['Message'])
      self.setApplicationStatus('Failed to setup MARLIN_DLL')
      return S_ERROR('Something wrong with software installation')
    marlin_dll = res["Value"]

    while True:
      res = self.cali.requestNewParameters()
      if not res['OK']:
        self.log.error('Stop executing calibration workflow. Error when requesting new parameters from the calibration'
                       ' service. Error message: %s' % res['Message'])
        return res

      while res['Value'] is None:
        self.log.notice("Waiting for new parameters set")
        wasteCPUCycles(10)
        res = self.cali.requestNewParameters()
        if not res['OK']:
          self.log.error('Stop executing calibration workflow. Error when requesting new parameters from the'
                         ' calibration service. Error message: %s' % res['Message'])
          return res

      calibrationParameters = res['Value']
      if calibrationParameters['calibrationIsFinished']:
        self.log.notice("Calibration finished")
        break

      self.currentPhase = calibrationParameters['currentPhase']
      self.currentStage = calibrationParameters['currentStage']
      self.currentStep = calibrationParameters['currentStep']
      parameterDict = calibrationParameters['parameters']

      # list of parameters which can be not present in the steering file be default
      # if they are not found - add them
      parametersToCheck = [('MaxClusterEnergyToApplySoftComp', 'float'), ('ECALLayers', 'IntVec')]
      for iParName, iParType in parametersToCheck:
        tmpKey = self.getKey(parameterDict, iParName)
        if tmpKey is not None:
          res = readParametersFromSteeringFile(self.SteeringFile, {tmpKey: None})
          if not res['OK']:
            #  expect key of following format:
            #    ".//processor[@name=%s]/parameter[@name='MaxClusterEnergyToApplySoftComp']"
            processorName = tmpKey.split('=')[1].split(']')[0]
            processorName = processorName.replace('\'', '')
            processorName = processorName.replace('\"', '')
            res = addParameterToProcessor(self.SteeringFile, processorName,
                                          {'name': iParName, 'type': iParType,
                                           'value': parameterDict[tmpKey]})

      self.setApplicationStatus('PandoraCalib_%s: stage: %s; phase: %s; step: %s' %
                                (self.calibrationID, self.currentStage, self.currentPhase, self.currentStep))

      res = self.resolveInputSlcioFilesAndAddToParameterDict(listofslcio, parameterDict)
      if res['OK']:
        parameterDict = res['Value']
      else:
        self.log.error('Problem while executing resolveInputSlcioFilesAndAddToParameterDict: %s' % res['Message'])
        return res

      steeringFileToRun = 'marlinSteeringFile_%s_%s_%s.xml' % (self.currentStage, self.currentPhase, self.currentStep)
      res = updateSteeringFile(self.SteeringFile, steeringFileToRun, parameterDict)
      if not res['OK']:
        self.log.error('Error while updateing steering file. Error message: %s' % res['Message'])
        return res

      # update path to the detector model, either local or from the software, if specified in settings
      if compactFile:
        res = updateSteeringFile(steeringFileToRun, steeringFileToRun,
                                 {".//processor[@name='InitDD4hep']/parameter[@name='DD4hepXMLFile']": compactFile})
        if not res['OK']:
          self.log.error('Error while updateing steering file. Error message: %s' % res['Message'])
          return res

      # FIXME for debug purposes:
      if (self.currentStage == 3) and (not os.path.exists('newPhotonLikelihood.xml')) and (self.currentStep == 0):
        # there is no newPhotonLikelihood.xml file -> create empty dummy and do not use it
        open('newPhotonLikelihood.xml', 'a').close()

      if self.currentStage == 3 and not os.path.exists('newPhotonLikelihood.xml'):
        newPhotonLikelihood = self.cali.requestNewPhotonLikelihood()
        while newPhotonLikelihood is None:
          self.log.notice("Waiting for new photon likelihood file for stage 3")
          wasteCPUCycles(10)
          newPhotonLikelihood = self.cali.requestNewPhotonLikelihood()
        stringToBinaryFile(newPhotonLikelihood, 'newPhotonLikelihood.xml')
        # TODO this depends from the name inside steering file... And it's even more difficult for FCCee case
        pandoraSettingsFile = readValueFromSteeringFile(
            self.SteeringFile, self.getKey(parameterDict, 'PandoraSettingsXmlFile'))
        pandoraSettingsFile = pandoraSettingsFile.strip()

        updateSteeringFile(
            pandoraSettingsFile, pandoraSettingsFile,
            {"algorithm[@type='PhotonReconstruction']/HistogramFile":
             "newPhotonLikelihood.xml"})

      # TODO clean up Marlin steering file - we don't need a lot of processors for calibration
      self.log.notice("new set of calibration parameters: %r" % parameterDict)

      self.result = self.runScript(steeringFileToRun, env_script_path, marlin_dll)
      if not self.result['OK']:
        self.log.error('Something wrong during running:', self.result['Message'])
        self.setApplicationStatus('Error during running %s' % self.applicationName)
        return S_ERROR('Failed to run %s' % self.applicationName)

      # FIXME make sure that runScript function return tuple of the same format as used below
      # self.result = {'OK':True,'Value':(0,'Disabled Execution','')}
      resultTuple = self.result['Value']
      if not os.path.exists(self.applicationLog):
        self.log.error("Something went terribly wrong, the log file is not present")
        self.setApplicationStatus('%s failed terribly, you are doomed!' % (self.applicationName))
        if not self.ignoreapperrors:
          return S_ERROR('%s did not produce the expected log' % (self.applicationName))

      # FIXME is result tuple correspond to the status?
      status = resultTuple
      # stdOutput = resultTuple[1]
      # stdError = resultTuple[2]
      self.log.info("Status after the application execution is:", str(status))

      outFile = "pfoAnalysis.root"
      if self.currentStage == 2:
        outFile = 'PandoraLikelihoodDataPhotonTraining.xml'

      self.cali.reportResult(outFile)

      self.setApplicationStatus('PandoraCalib_%s: step %s is finished. Waiting for other jobs' %
                                (self.calibrationID, self.currentStep))

    # TODO implement me
    #  return self.finalStatusReport(status)

    self.setApplicationStatus('PandoraCalib_%s: Calibration is finished.' % self.calibrationID)
    return S_OK()

  def getKey(self, parameterDict, pattern):
    """Return key from parameterDict which satisfy pattern."""
    for iKey in parameterDict:
      if pattern in iKey:
        return iKey
    self.log.error('Cannot find XPath inside the parameter dict which contains pattern: %s' % pattern)
    #  try:
    #    if pattern in iKey:
    #      return iKey
    #  except:
    #    self.log.error('Cannot find XPath inside the parameter dict which contains pattern: %s' % pattern)
    return None

  def resolveInputSlcioFilesAndAddToParameterDict(self, allSlcioFiles, parameterDict):
    """Resolve pathes to input files.

    Add PandoraSettings-file and input slcio files which corresponds to current currentStage and currentPhase to the
    parameterDict.

    :param list str allSlcioFiles: List of all slcio-files in the node
    :param dict parameterDict: dict of parameters and their values

    :returns: S_OK or S_ERROR
    :rtype: dict
    """
    pandoraSettingsFile = ''
    if self.currentStage in [1, 3]:  # FIXME hardcoded values are bad...
      pandoraSettingsFile = 'PandoraSettings/PandoraSettingsDefault.xml'
      if 'FCC' in self.detectorModel:  # pylint: disable=unsupported-membership-test
        pandoraSettingsFile = 'PandoraSettingsFCCee/PandoraSettingsDefault.xml'
    else:
      pandoraSettingsFile = 'CalibrationPandoraSettings/PandoraSettingsPhotonTraining.xml'

    iType = CalibrationPhase.fileKeyFromPhase(self.currentPhase).lower()
    self.log.info('SASHA iType: %s' % iType)
    res = self.cali.getInputDataDict()
    self.log.info('getInputDataDict:', ' %s' % res)
    if not res['OK']:
      errorMessageConst = 'Somemething went wrong during retrieveing inputDataDict!'
      errorMessageVariable = 'Msg: %s' % (res['Message'])
      self.log.error(errorMessageConst, errorMessageVariable)
      return S_ERROR(errorMessageConst + errorMessageVariable)

    self.log.info('SASHA self.cali.getInputDataDict() %s' % res)

    inputDataDict = res['Value']
    self.log.debug('iType:', ' %s' % iType)
    self.log.debug('inputDataDict.keys():', ' %s' % list(inputDataDict.keys()))
    if iType not in inputDataDict:
      errorMessage = 'Corrupted inputDataDict! No files for process: %s' % (iType)
      self.log.error(errorMessage)
      return S_ERROR(errorMessage)

    lfnList = inputDataDict[iType][0]
    numberOfEventsToSkip = inputDataDict[iType][1]
    numberOfEventsToProcess = inputDataDict[iType][2]

    parameterDict[self.getKey(parameterDict, 'SkipNEvents')] = numberOfEventsToSkip
    parameterDict[self.getKey(parameterDict, 'MaxRecordNumber')] = numberOfEventsToProcess

    res = resolveIFpaths(lfnList)
    if not res['OK']:
      self.log.error("Failed to resolve path to input slcio files: %s" % res)
      return res
    filesToRunOn = res['Value']
    if len(filesToRunOn) == 0:
      errorMessage = 'Corrupted inputDataDict! No files for process: %s' % (iType)
      self.log.error(errorMessage)
      return S_ERROR(errorMessage)
    if not set(filesToRunOn).issubset(allSlcioFiles):
      errorMessage = ('Cannot find all input data on the worker. Needed files for phase: %s; All copied files: %s'
                      % (filesToRunOn, allSlcioFiles))
      self.log.error(errorMessage)
      return S_ERROR(errorMessage)

    parameterDict[self.getKey(parameterDict, 'LCIOInputFiles')] = ' '.join(filesToRunOn)
    parameterDict[self.getKey(parameterDict, 'PandoraSettingsXmlFile')] = pandoraSettingsFile

    # TODO should one use different steering file for photon training? if no one need to append line below during all
    # steps
    if self.currentStage in [1, 3]:
      parameterDict[self.getKey(parameterDict, 'RootFile')] = 'pfoAnalysis.root'
    else:
      parameterDict[self.getKey(parameterDict, 'RootFile')] = 'dummy.root'
    return S_OK(parameterDict)

  def runScript(self, marlinSteeringFile, env_script_path, marlin_dll):
    """Actual bit of code running Marlin and PandoraAnalysis.

    :param marlinSteeringFile: steering file to use for Marlin reconstruction. E.g.: 'fccReconstruction.xml'
    :param str env_script_path: path to the setup environment scripts
    :param str marlin_dll: string containing path to marlin libraries

    :returns: FIXME S_OK or S_ERROR
    :rtype: dict
    """
    res = self._prepareRunScript(marlinSteeringFile, env_script_path, marlin_dll)
    if not res['OK']:
      return res

    scriptName = res['Value']

    if os.path.exists(self.applicationLog):
      os.remove(self.applicationLog)

    os.chmod(scriptName, 0o755)
    comm = 'sh -c "./%s"' % (scriptName)
    self.setApplicationStatus('Running: stage: %s; phase: %s; step: %s'
                              % (self.currentStage, self.currentPhase, self.currentStep))
    self.stdError = ''
    res = shellCall(0, comm, callbackFunction=self.redirectLogOutput, bufferLimit=20971520)
    return res

  def _prepareRunScript(self, marlinSteeringFile, env_script_path, marlin_dll):
    """Return current parameters.

    :param marlinSteeringFile: steering file to use for Marlin reconstruction. E.g.: 'fccReconstruction.xml'
    :param str env_script_path: path to the setup environment scripts
    :param str marlin_dll: string containing path to marlin libraries

    :returns: S_OK or S_ERROR
    :rtype: dict
    """
    scriptName = '%s_%s_Run_%s.sh' % (self.applicationName, self.applicationVersion, self.currentStep)
    if os.path.exists(scriptName):
      os.remove(scriptName)
    script = open(scriptName, 'w')
    script.write('#!/bin/bash \n')
    script.write('#####################################################################\n')
    script.write('# Dynamically generated script to run a production or analysis job. #\n')
    script.write('#####################################################################\n')
    script.write("source %s\n" % env_script_path)
    script.write("declare -x MARLIN_DLL=%s\n" % marlin_dll)
    if os.path.exists("./lib/lddlib"):
      script.write('declare -x LD_LIBRARY_PATH=./lib/lddlib:$LD_LIBRARY_PATH\n')
    script.write('declare -x PATH=$ROOTSYS/bin:$PATH\n')
    script.write('declare -x MARLIN_DEBUG=1\n')  # Needed for recent version of marlin (from 03 april 2013)
    # We need to make sure the PandoraSettings is in the current directory
    script.write("""
if [ -e "${PANDORASETTINGS}" ]
then
   cp $PANDORASETTINGS .
fi
""")
    script.write('echo =============================\n')
    script.write('echo LD_LIBRARY_PATH is\n')
    script.write('echo $LD_LIBRARY_PATH | tr ":" "\n"\n')
    script.write('echo =============================\n')
    script.write('echo PATH is\n')
    script.write('echo $PATH | tr ":" "\n"\n')
    script.write('echo =============================\n')
    script.write('echo MARLIN_DLL is\n')
    script.write('echo $MARLIN_DLL | tr ":" "\n"\n')
    script.write('echo =============================\n')
    if self.debug:
      script.write('echo ldd of executable is\n')
      script.write('ldd `which Marlin` \n')
      script.write('echo =============================\n')
      if os.path.exists('./lib/marlin_dll'):
        script.write('ldd ./lib/marlin_dll/*.so \n')
      if os.path.exists('./lib/lddlib'):
        script.write('ldd ./lib/lddlib/*.so \n')
      script.write('echo =============================\n')
    script.write('env | sort >> localEnv.log\n')

    if not os.path.exists(marlinSteeringFile):
      script.close()
      self.log.error("Steering file missing: %s" % (marlinSteeringFile))
      return S_ERROR("SteeringFile is missing: %s" % (marlinSteeringFile))
    # check
    script.write('Marlin -c %s %s\n' % (marlinSteeringFile, self.extraCLIarguments))
    # real run
    script.write('Marlin %s %s\n' % (marlinSteeringFile, self.extraCLIarguments))
    script.write('declare -x appstatus=$?\n')
    script.write('exit $appstatus\n')
    script.close()
    return S_OK(scriptName)



#!/usr/local/env python
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test Marlin module."""

from __future__ import absolute_import
import inspect
import unittest
from mock import patch, MagicMock as Mock

from DIRAC import S_OK, S_ERROR
from ILCDIRAC.Interfaces.API.NewInterface.Applications import Marlin
from Tests.Utilities.GeneralUtils import assertEqualsImproved, assertDiracFailsWith, \
    assertDiracSucceeds

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Interfaces.API.NewInterface.Applications.Marlin'

# pylint: disable=protected-access


class MarlinTestCase(unittest.TestCase):
  """Base class for the Marlin test cases."""

  def setUp(self):
    """set up the objects."""
    self.mar = Marlin({})

  def test_setgear(self):
    self.mar.setGearFile('lfn:/my/gear/file.txt')
    self.assertFalse(self.mar._errorDict)
    self.assertIn('lfn:/my/gear/file.txt', self.mar.inputSB)

  def test_setKeepRecFile(self):
    """Tests for behaviour with KeepRecFile True/False."""
    self.assertTrue(self.mar.keepRecFile)
    self.mar.setKeepRecFile(False)
    self.assertFalse(self.mar._errorDict)
    self.mar.setKeepRecFile(True)
    self.assertTrue(self.mar.keepRecFile)
    self.assertFalse(self.mar._errorDict)
    self.mar.setKeepRecFile(123)
    self.assertIn('val = 123', str(self.mar._errorDict))

  def test_setoutputrec(self):
    self.mar.setOutputRecFile('my/file.outputrec', 'mytestPath')
    assertEqualsImproved(self.mar.outputRecPath, 'mytestPath', self)
    self.assertFalse(self.mar._errorDict)

  def test_setoutputdst(self):
    self.mar.setOutputDstFile('my/file.outputdst', 'mytestPath')
    assertEqualsImproved(self.mar.outputDstPath, 'mytestPath', self)
    self.assertFalse(self.mar._errorDict)

  def test_setproclist(self):
    self.mar.setProcessorsToUse(['proc1', 'proc2'])
    self.assertFalse(self.mar._errorDict)

  def test_setexcludeproclist(self):
    self.mar.setProcessorsToExclude(['proc1', 'proc2'])
    self.assertFalse(self.mar._errorDict)

  def test_userjobmodules(self):
    module_mock = Mock()
    assertDiracSucceeds(self.mar._userjobmodules(module_mock), self)

  def test_prodjobmodules(self):
    module_mock = Mock()
    assertDiracSucceeds(self.mar._prodjobmodules(module_mock), self)

  def test_prodjobmodules_outputpath(self):
    module_mock = Mock()
    self.mar.outputPath = 'aef'
    assertDiracSucceeds(self.mar._prodjobmodules(module_mock), self)
    self.assertIn({'OutputFile': '@{OutputFile}', 'outputPath': '@{OutputPath}',
                     'outputDataSE': '@{OutputSE}'}, self.mar._listofoutput)

  def test_userjobmodules_fails(self):
    with patch('%s._setUserJobFinalization' % MODULE_NAME, new=Mock(return_value=S_OK('something'))),\
        patch('%s._setApplicationModuleAndParameters' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_test_err'))):
      assertDiracFailsWith(self.mar._userjobmodules(None),
                            'userjobmodules failed', self)

  def test_prodjobmodules_fails(self):
    with patch('%s._setApplicationModuleAndParameters' % MODULE_NAME, new=Mock(return_value=S_OK('something'))), \
        patch('%s._setOutputComputeDataList' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_other_test_err'))):
      assertDiracFailsWith(self.mar._prodjobmodules(None),
                            'prodjobmodules failed', self)

  def test_checkconsistency(self):
    self.mar.version = '13'
    self.mar.steeringFile = '/mysteer/file.stdhep'
    self.mar.gearFile = None
    self.mar._jobtype = 'notUser'
    self.mar.outputFile = None
    with patch('os.path.exists', new=Mock(return_value=True)), \
         patch.object(inspect.getmodule(Marlin), 'checkXMLValidity', new=Mock(return_value=S_OK())):
      assertDiracSucceeds(self.mar._checkConsistency(), self)
      self.assertIn({'outputFile': '@{outputREC}', 'outputPath': '@{outputPathREC}',
                       'outputDataSE': '@{OutputSE}'}, self.mar._listofoutput)
      self.assertIn({'outputFile': '@{outputDST}', 'outputPath': '@{outputPathDST}',
                       'outputDataSE': '@{OutputSE}'}, self.mar._listofoutput)
      for keyword in ['detectorType', 'marlin_gearfile', 'marlin_steeringfile']:
        self.assertIn(keyword, self.mar.prodparameters)
      assertEqualsImproved(self.mar.gearFile, None, self)

  def test_checkconsistency_noversion(self):
    self.mar.version = None
    assertDiracFailsWith(self.mar._checkConsistency(), 'version not set', self)

  def test_checkconsistency_invalidxml(self):
    self.mar.version = '13'
    self.mar.steeringFile = '/mysteer/file.stdhep'
    with patch('os.path.exists', new=Mock(return_value=True)), \
         patch.object(inspect.getmodule(Marlin), 'checkXMLValidity', new=Mock(return_value=S_ERROR('mytesterrxml'))):
      assertDiracFailsWith(self.mar._checkConsistency(), 'supplied steering file cannot be read with xml', self)

  def test_checkconsistency_othercase(self):
    self.mar.version = '13'
    self.mar.steeringFile = '/mysteer/file.stdhep'
    self.mar.gearFile = 'myGearOutput.mock'
    self.mar._jobtype = 'notUser'
    self.mar.outputFile = 'myoutput.test'
    with patch('os.path.exists', new=Mock(return_value=False)):
      assertDiracSucceeds(self.mar._checkConsistency(), self)
      self.assertNotIn({'outputFile': '@{outputREC}', 'outputPath': '@{outputPathREC}',
                          'outputDataSE': '@{OutputSE}'}, self.mar._listofoutput)
      self.assertNotIn({'outputFile': '@{outputDST}', 'outputPath': '@{outputPathDST}',
                          'outputDataSE': '@{OutputSE}'}, self.mar._listofoutput)
      for keyword in ['detectorType', 'marlin_gearfile', 'marlin_steeringfile']:
        self.assertIn(keyword, self.mar.prodparameters)
      assertEqualsImproved(self.mar.gearFile, 'myGearOutput.mock', self)

  def test_checkconsistency_lastcase(self):
    self.mar.version = '13'
    self.mar.steeringFile = None
    self.mar.gearFile = 'myGearOutput.mock'
    self.mar._jobtype = 'User'
    self.mar.outputFile = 'myoutput.test'
    assertDiracSucceeds(self.mar._checkConsistency(), self)
    self.assertNotIn({'outputFile': '@{outputREC}', 'outputPath': '@{outputPathREC}',
                        'outputDataSE': '@{OutputSE}'}, self.mar._listofoutput)
    self.assertNotIn({'outputFile': '@{outputDST}', 'outputPath': '@{outputPathDST}',
                        'outputDataSE': '@{OutputSE}'}, self.mar._listofoutput)
    for keyword in ['detectorType', 'marlin_gearfile', 'marlin_steeringfile']:
      self.assertNotIn(keyword, self.mar.prodparameters)
    assertEqualsImproved(self.mar.gearFile, 'myGearOutput.mock', self)

  def test_resolvelinkedparams(self):
    """Test _resolveLinkedStepParameters with something happening."""
    step_mock = Mock()
    input_mock = Mock()
    input_mock.getType.return_value = {'abc': False}
    self.mar._linkedidx = 3
    self.mar._jobsteps = [None, None, None, input_mock]
    assertDiracSucceeds(self.mar._resolveLinkedStepParameters(step_mock), self)
    step_mock.setLink.assert_called_once_with('InputFile', {'abc': False}, 'OutputFile')

  def test_resolvelinkedparams_noinputstep(self):
    """Call _resolveLinkedStep function, which does nothing."""
    self.mar._linkedidx = None
    self.mar._inputappstep = []
    assertDiracSucceeds(self.mar._resolveLinkedStepParameters(None), self)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Mokka analysis module.

Called by Job Agent.

:since:  Jan 29, 2010

:author: Stephane Poss
:author: Przemyslaw Majewski
"""
from __future__ import absolute_import
import os
import shutil

from DIRAC.Core.Utilities.Subprocess import shellCall
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC import S_OK, S_ERROR, gLogger

from ILCDIRAC.Workflow.Modules.ModuleBase import ModuleBase, generateRandomString
from ILCDIRAC.Core.Utilities.CombinedSoftwareInstallation import getSoftwareFolder, getEnvironmentScript
from ILCDIRAC.Core.Utilities.PrepareOptionFiles import prepareSteeringFile, getNewLDLibs
from ILCDIRAC.Core.Utilities.PrepareLibs import removeLibc
from ILCDIRAC.Core.Utilities.resolvePathsAndNames import resolveIFpaths, getProdFilename
from ILCDIRAC.Core.Utilities.FindSteeringFileDir import getSteeringFileDirName

__RCSID__ = '$Id$'
LOG = gLogger.getSubLogger(__name__)


class MokkaAnalysis(ModuleBase):
  """Specific Module to run a Mokka job."""

  def __init__(self):
    super(MokkaAnalysis, self).__init__()
    self.enable = True
    self.STEP_NUMBER = ''
    self.InputFile = []
    self.macFile = ''
    self.detectorModel = ''
    self.run_number = 0
    self.firstEventNumber = 1
    self.applicationName = 'Mokka'
    self.dbSlice = ''
    self.startFrom = 0
    self.eventstring = ['>>> Event']
    self.ProcessID = ''
    self.RandomSeed = 0
    self.mcRunNumber = 0
    self.db_dump_name = ""


#############################################################################

  def applicationSpecificInputs(self):
    """Resolve all input variables for the module here.

    :return: S_OK()
    """

    if self.WorkflowStartFrom:
      self.startFrom = self.WorkflowStartFrom

    if 'stdhepFile' in self.step_commons:
      inputf = self.step_commons["stdhepFile"]
      if not isinstance(inputf, list):
        inputf = inputf.split(";")
      self.InputFile = inputf

    self.RandomSeed = self.determineRandomSeed()

    resDBSlice = self.determineDBSlice()
    if not resDBSlice['OK']:
      return resDBSlice

    if "IS_PROD" in self.workflow_commons:
      if self.workflow_commons["IS_PROD"]:
        self.mcRunNumber = self.RandomSeed
        # self.OutputFile = getProdFilename(self.outputFile,int(self.workflow_commons["PRODUCTION_ID"]),
        #                                  int(self.workflow_commons["JOB_ID"]))
        if 'ProductionOutputData' in self.workflow_commons:
          outputlist = self.workflow_commons['ProductionOutputData'].split(";")
          for obj in outputlist:
            if obj.lower().count("_sim_"):
              self.OutputFile = os.path.basename(obj)
            elif obj.lower().count("_gen_"):
              self.InputFile = [os.path.basename(obj)]
        else:
          self.OutputFile = getProdFilename(self.OutputFile, int(self.workflow_commons["PRODUCTION_ID"]),
                                            int(self.workflow_commons["JOB_ID"]),
                                            self.workflow_commons,
                                           )
          # if 'WhizardOutput' in self.workflow_commons:
          #  self.InputFile = getProdFilename(self.workflow_commons["WhizardOutput"],
          #                                    int(self.workflow_commons["PRODUCTION_ID"]),
          #                                    int(self.workflow_commons["JOB_ID"]))
          self.InputFile = [getProdFilename(self.InputFile, int(self.workflow_commons["PRODUCTION_ID"]),
                                            int(self.workflow_commons["JOB_ID"]))]

    if len(self.InputData):
      if 'EvtClass' in self.inputdataMeta and not self.ProcessID:
        self.ProcessID = self.inputdataMeta['EvtClass']
      if 'EvtType' in self.inputdataMeta and not self.ProcessID:
        self.ProcessID = self.inputdataMeta['EvtType']

    if not len(self.InputFile) and len(self.InputData):
      for files in self.InputData:
        if files.lower().find(".stdhep") > -1 or files.lower().find(".hepevt") > -1:
          self.InputFile.append(files)
          break
    return S_OK('Parameters resolved')

  def runIt(self):
    """Called by ModuleBase.

    Executes the following:
      - read the application parameters that where defined in ILCJob, and stored in the job definition
      - setup the SQL server and run it in the background
      - prepare the steering file using :any:`prepareSteeringFile`
      - run Mokka and catch its return status

    :return: S_OK(), S_ERROR()
    """

    # if not self.workflowStatus['OK'] or not self.stepStatus['OK']:
    #  LOG.info('Skip this module, failure detected in a previous step :')
    #  LOG.info('Workflow status : %s' %(self.workflowStatus))
    #  LOG.info('Step Status %s' %(self.stepStatus))
    #  return S_OK()

    self.result = S_OK()

    if not self.platform:
      self.result = S_ERROR('No ILC platform selected')
    elif not self.applicationLog:
      self.result = S_ERROR('No Log file provided')

    if not self.result['OK']:
      LOG.error("Failed to resolve the input parameters:", self.result["Message"])
      return self.result

    if not self.workflowStatus['OK'] or not self.stepStatus['OK']:
      LOG.verbose('Workflow status = %s, step status = %s' % (self.workflowStatus['OK'], self.stepStatus['OK']))
      return S_OK('Mokka should not proceed as previous step did not end properly')

    LOG.info("Executing Mokka %s" % (self.applicationVersion))

    if self.dbSlice:
      if self.dbSlice.lower().startswith("lfn"):
        self.dbSlice = os.path.basename(self.dbSlice)
      self.db_dump_name = self.dbSlice
    else:
      self.db_dump_name = "CLICMokkaDB.sql"

    res = getEnvironmentScript(self.platform, "mokka", self.applicationVersion, self.getEnvScript)
    LOG.notice("Got the environment script: %s" % res)
    if not res['OK']:
      LOG.error("Error getting the env script: ", res['Message'])
      return res
    env_script_path = res['Value']

    # Setup MySQL instance
    mysqlBasePath = '%s/mysqltmp/MokkaDBRoot-%s' % (os.getcwd(), generateRandomString(8))
    LOG.notice("Placing mysql files in %s" % mysqlBasePath)

    # steering file that will be used to run
    mokkasteer = "mokka.steer"
    if os.path.exists("mokka.steer"):
      try:
        os.rename("mokka.steer", "mymokka.steer")
      except EnvironmentError as err:
        LOG.error("Failed renaming the steering file: ", str(err))
      self.SteeringFile = "mymokka.steer"

    # prepare steering file
    # first, I need to take the stdhep file, find its path (possible LFN)
    if len(self.InputFile) > 0:
      #self.InputFile = os.path.basename(self.InputFile)
      res = resolveIFpaths(self.InputFile)
      if not res['OK']:
        LOG.error("Generator file not found")
        return res
      self.InputFile = res['Value']
    if len(self.macFile) > 0:
      self.macFile = os.path.basename(self.macFile)
    # idem for steering file

    self.SteeringFile = os.path.basename(self.SteeringFile)
    if not os.path.exists(self.SteeringFile):
      LOG.verbose("Steering file %s not found locally" % self.SteeringFile)
      res = getSteeringFileDirName(self.platform, "mokka", self.applicationVersion)
      if not res['OK']:
        LOG.error("Missing Steering file directory:", res['Message'])
        return res
      steeringfiledirname = res['Value']
      if os.path.exists(os.path.join(steeringfiledirname, self.SteeringFile)):
        try:
          shutil.copy(os.path.join(steeringfiledirname, self.SteeringFile), "./" + self.SteeringFile)
        except EnvironmentError as err:
          LOG.error("Failed copying file", self.SteeringFile)
          return S_ERROR('Failed to access file %s: %s' % (self.SteeringFile, str(err)))
          #self.steeringFile = os.path.join(mySoftwareRoot,"steeringfiles",self.steeringFile)
    if not os.path.exists(self.SteeringFile):
      LOG.error("Missing steering file, should not happen!")
      return S_ERROR("Could not find steering file")
    else:
      LOG.verbose("Found local copy of %s" % self.SteeringFile)
    # The following is because if someone uses particle gun, there is no InputFile
    if not len(self.InputFile):
      self.InputFile = ['']
    steerok = prepareSteeringFile(self.SteeringFile, mokkasteer, self.detectorModel, self.InputFile[0],
                                  self.macFile, self.NumberOfEvents, self.startFrom, self.RandomSeed,
                                  self.mcRunNumber,
                                  self.ProcessID,
                                  self.debug,
                                  self.OutputFile,
                                  self.inputdataMeta)
    if not steerok['OK']:
      LOG.error('Failed to create MOKKA steering file')
      return S_ERROR('Failed to create MOKKA steering file')

    # Extra option depending on mokka version
    mokkaextraoption = ""
    if self.applicationVersion not in ["v07-02", "v07-02fw", "v07-02fwhp", "MokkaRevision42", "MokkaRevision43",
                                       "MokkaRevision44", "Revision45"]:
      mokkaextraoption = "-U"

    scriptName = 'Mokka_%s_Run_%s.sh' % (self.applicationVersion, self.STEP_NUMBER)

    if os.path.exists(scriptName):
      os.remove(scriptName)
    script = open(scriptName, 'w')
    script.write('#!/bin/bash \n')
    script.write('#####################################################################\n')
    script.write('# Dynamically generated script to run a production or analysis job. #\n')
    script.write('#####################################################################\n')
    script.write("source %s\n" % env_script_path)
    if os.path.exists("./lib"):
      script.write('declare -x LD_LIBRARY_PATH=./lib:$LD_LIBRARY_PATH\n')
    script.write('echo =============================\n')
    script.write('echo Content of mokka.steer:\n')
    script.write('cat mokka.steer\n')
    if self.macFile:
      script.write('echo =============================\n')
      script.write('echo Content of %s\n' % os.path.basename(self.macFile))
      script.write("cat %s\n" % os.path.basename(self.macFile))
    elif os.path.exists("./mokkamac.mac"):
      script.write("""
echo =============================
echo Content of mokkamac.mac:
cat mokkamac.mac
""")
    script.write('echo =============================\n')
    script.write('echo LD_LIBRARY_PATH is\n')
    script.write('echo $LD_LIBRARY_PATH | tr ":" "\n"\n')
    script.write('echo =============================\n')
    script.write('echo PATH is\n')
    script.write('echo $PATH | tr ":" "\n"\n')
    if self.debug:
      script.write('echo ldd of Mokka is\n')
      script.write('ldd `which Mokka` \n')
      script.write('echo =============================\n')
      script.write('echo ldd of mysql is\n')
      script.write('ldd `which mysql` \n')
      script.write('echo =============================\n')
    script.write('env | sort >> localEnv.log\n')
    script.write('echo =============================\n')
    script.write("""if [ ! -e ./%(DBDUMP)s ]; then
    echo "Missing DB dump!" >&2
    exit 1
fi
    """ % {'DBDUMP': self.db_dump_name})
    # Now start the MySQL server and configure.
    script.write("declare -x MOKKADBROOT=%s\n" % mysqlBasePath)
    script.write("declare -x WORKDIR=%s\n" % os.getcwd())
    script.write("declare -x LOGDIR=$WORKDIR/mysqllogs/\n")
    script.write("mkdir -p $LOGDIR\n")

    # mysql socket must not be longer than 107 characters!
    script.write("declare -x SOCKETBASE=/tmp/mokka-%s\n" % generateRandomString(8))
    script.write("mkdir -p $SOCKETBASE\n")
    script.write("declare -x SOCKETPATH=$SOCKETBASE/mysql.sock\n")
    script.write("mkdir -p $MOKKADBROOT\n")
    script.write("declare -x MYSQLDATA=$MOKKADBROOT/data\n")
    script.write("rm -rf $MYSQLDATA\n")
    script.write("mkdir -p $MYSQLDATA\n")
    script.write("date\n")
    script.write("""echo "*** Installing MySQL"
echo "mysql_install_db --no-defaults --skip-networking --socket=$SOCKETPATH --datadir=$MYSQLDATA --basedir=$MYSQL --pid-file=$MOKKADBROOT/mysql.pid --log-error=$LOGDIR/mysql.err --log=$LOGDIR/mysql.log"
mysql_install_db --no-defaults --skip-networking --socket=$SOCKETPATH --datadir=$MYSQLDATA --basedir=$MYSQL --pid-file=$MOKKADBROOT/mysql.pid --log-error=$LOGDIR/mysql.err --log=$LOGDIR/mysql.log
install_st=$?
if [ $install_st -ne 0 ]
then
      exit $install_st
fi
date
echo "*** Running mysqld-safe"
cd $MYSQL
bin/mysqld_safe --no-defaults --skip-networking --socket=$SOCKETPATH --datadir=$MYSQLDATA --basedir=$MYSQL --pid-file=$MOKKADBROOT/mysql.pid --log-error=$LOGDIR/mysql.err --log=$LOGDIR/mysql.log &
COUNT=0
while [ -z "$socket_grep" ] ; do
    socket_grep=$(netstat -ln 2>/dev/null | grep "$SOCKETPATH")
    echo -n .
    sleep 1
    if [ $COUNT -eq 100 ]; then
        echo "Failed to find socket"
        break
    fi
    COUNT=$(( $COUNT + 1 ))
done
cd -
echo "*** Configuring MySQL"
mysqladmin --no-defaults -hlocalhost --socket=$SOCKETPATH -uroot password 'rootpass'
declare -x admin_st=$?
if [ $admin_st -ne 0 ]
then
  echo "*** mysqladmin failed, cannot proceed" >&2
  mysqladmin --no-defaults -hlocalhost --socket=$SOCKETPATH -uroot -prootpass shutdown
  exit $admin_st
fi

mysql --no-defaults -uroot -hlocalhost --socket=$SOCKETPATH -prootpass <<< 'GRANT ALL PRIVILEGES ON *.* TO root;'
mysql --no-defaults -uroot -hlocalhost --socket=$SOCKETPATH -prootpass <<< 'GRANT ALL PRIVILEGES ON *.* TO consult IDENTIFIED BY \"consult\";'
mysql --no-defaults -uroot -hlocalhost --socket=$SOCKETPATH -prootpass <<< 'DELETE FROM mysql.user WHERE User = \"\"; FLUSH PRIVILEGES;'
echo "*** Installing Mokka DB"
date
echo "mysql --no-defaults -hlocalhost --socket=$SOCKETPATH -uroot -prootpass < ./%(DBDUMP)s"
time mysql --no-defaults -hlocalhost --socket=$SOCKETPATH -uroot -prootpass < ./%(DBDUMP)s
sleep 5
date\n""" % {'DBDUMP': self.db_dump_name})
    # Now take care of the particle tables.
    script.write("""
if [ -e "./particle.tbl" ]
then
   declare -x PARTICLETBL=./particle.tbl
fi
if [ -d "${MOKKA}" ]
#As the local env does not set MOKKA, it must come from CVMFS
then
   declare -x PARTICLETBL=$MOKKA/particle.tbl
fi
if [ -d "${MOKKATARBALL}" ]
#As the local env does set MOKKATARBALL, it must be there
then
   declare -x PARTICLETBL=$MOKKATARBALL/ConfigFiles/particle.tbl
fi
echo "/Mokka/init/PDGFile $PARTICLETBL" >> %s
\n""" % mokkasteer)

    # Now run Mokka
    comm = 'Mokka %s -h localhost:$SOCKETPATH %s %s\n' % (mokkaextraoption, mokkasteer, self.extraCLIarguments)
    LOG.info("Command : %s" % (comm))
    script.write(comm)
    script.write('declare -x appstatus=$?\n')

    # Now shutdown the MySQL server
    script.write("mysqladmin --no-defaults -hlocalhost --socket=$SOCKETPATH -uroot -prootpass shutdown\n")
    script.write("""
while [ -n "$socket_grep" ] ; do
    socket_grep=$(netstat -ln 2>/dev/null | grep "$SOCKETPATH")
    echo -n .
    sleep 1
done
""")
    script.write("rm -f %s\n" % self.db_dump_name)  # remove db file
    script.write('rm -rf $MYSQLDATA\n')  # cleanup
    script.write('rm -rf $MOKKADBROOT\n')  # cleanup
    # script.write('quit\n')
    # script.write('EOF\n')

    script.write('exit $appstatus\n')

    script.close()
    if os.path.exists(self.applicationLog):
      os.remove(self.applicationLog)

    os.chmod(scriptName, 0o755)
    comm = 'sh -c "./%s"' % scriptName
    self.setApplicationStatus('Mokka %s step %s' % (self.applicationVersion, self.STEP_NUMBER))
    self.stdError = ''
    self.result = shellCall(0, comm, callbackFunction=self.redirectLogOutput, bufferLimit=20971520)
    #self.result = {'OK':True,'Value':(0,'Disabled Execution','')}
    resultTuple = self.result['Value']

    status = resultTuple[0]
    # stdOutput = resultTuple[1]
    # stdError = resultTuple[2]
    LOG.info("Status after Mokka execution is %s" % str(status))
    if not os.path.exists(self.applicationLog):
      LOG.error("Something went terribly wrong, the log file is not present")
      self.setApplicationStatus('%s failed terribly, you are doomed!' % (self.applicationName))
      if not self.ignoreapperrors:
        LOG.error("Missing log file")
        return S_ERROR('%s did not produce the expected log' % (self.applicationName))
    # Now change the name of Mokka output to the specified filename
    if os.path.exists("out.slcio"):
      if len(self.OutputFile) > 0:
        os.rename("out.slcio", self.OutputFile)

    failed = False
    if status not in [0, 106, 9]:
      LOG.error("Mokka execution completed with errors:")
      failed = True
    elif status in [106, 9]:
      LOG.info("Mokka execution reached end of input generator file")
    else:
      LOG.info("Mokka execution finished successfully")

    message = 'Mokka %s Successful' % (self.applicationVersion)
    if failed is True:
      LOG.error("==================================\n StdError:\n")
      LOG.error(self.stdError)
      #self.setApplicationStatus('%s Exited With Status %s' %(self.applicationName,status))
      LOG.error('Mokka Exited With Status %s' % (status))
      message = 'Mokka Exited With Status %s' % (status)
      self.setApplicationStatus(message)
      if not self.ignoreapperrors:
        LOG.error("Application exists with error:", message)
        return S_ERROR(message)
    else:
      if status in [106, 9]:
        message = 'Mokka %s reached end of input generator file' % (self.applicationVersion)
      self.setApplicationStatus(message)
    return S_OK({'OutputFile': self.OutputFile})

  def getEnvScript(self, sysconfig, appname, appversion):
    """Return the environment script in case CVMFS is not here."""
    env_script_name = "MokkaEnv.sh"

    res = getSoftwareFolder(sysconfig, appname, appversion)
    if not res['OK']:
      LOG.error("Mokka: could not find installation directory!")
      return res
    myMokkaDir = res['Value']

    mySoftwareRoot = os.sep.join(myMokkaDir.rstrip(os.sep).split(os.sep)[0:-1])
    # Need to fetch the new LD_LIBRARY_PATH
    new_ld_lib_path = getNewLDLibs(self.platform, "mokka", self.applicationVersion)

    # Remove libc
    removeLibc(myMokkaDir)

    remoteMysqlInstall = mySoftwareRoot
    removeLibc(remoteMysqlInstall + '/mysql4grid/lib64/mysql')

    script = open(env_script_name, "w")
    script.write('#!/bin/sh \n')
    script.write('######################################################\n')
    script.write('# Dynamically generated script to get teh Mokka env. #\n')
    script.write('######################################################\n')
    # if(os.path.exists(sharedArea+"/initILCSOFT.sh")):
    #    script.write("%s/initILCSOFT.sh"%sharedArea)
    script.write("declare -x g4releases=%s\n" % (myMokkaDir))
    script.write("declare -x G4SYSTEM=Linux-g++\n")
    script.write("declare -x G4INSTALL=$g4releases/share/$g4version\n")
    #script.write("export G4SYSTEM G4INSTALL G4LIB CLHEP_BASE_DIR\n")
    script.write('declare -x G4LEDATA="$g4releases/G4LEDATA"\n')
    script.write('declare -x G4NEUTRONHPDATA="$g4releases/sl4/g4data/g4dataNDL"\n')
    script.write('declare -x G4LEVELGAMMADATA="$g4releases/sl4/g4data/g4dataPhotonEvap"\n')
    script.write('declare -x G4RADIOACTIVEDATA="$g4releases/sl4/g4data/g4dataRadiativeDecay"\n')
    # No such data on the GRID (???)
    # script.write('G4ELASTICDATA="$g4releases/share/data/G4ELASTIC1.1"\n')
    script.write('declare -x G4ABLADATA="$g4releases/sl4/g4data/g4dataABLA"\n')
    #script.write("export G4LEDATA G4NEUTRONHPDATA G4LEVELGAMMADATA G4RADIOACTIVEDATA G4ABLADATA\n")
    script.write('declare -x G4NEUTRONHP_NEGLECT_DOPPLER=1\n')
    script.write('declare -x PATH=%s:$PATH\n' % myMokkaDir)
    script.write('declare -x LD_LIBRARY_PATH=%s\n' % (myMokkaDir))
    script.write('declare -x MYSQL=%s/mysql4grid\n' % (remoteMysqlInstall))
    script.write('declare -x MOKKATARBALL=%s\n' % myMokkaDir)
    if new_ld_lib_path:
      script.write('declare -x LD_LIBRARY_PATH=%s:$LD_LIBRARY_PATH\n' % (new_ld_lib_path))

    script.write('declare -x PATH=%s/mysql4grid/bin:$PATH\n' % remoteMysqlInstall)  # for MySQL
    script.write('declare -x LD_LIBRARY_PATH=%s/mysql4grid/lib64/mysql:$LD_LIBRARY_PATH\n' % remoteMysqlInstall)
    # for MySQL
    script.write("if [ -e %s/%s ]; then\n" % (myMokkaDir, self.db_dump_name))
    script.write("  cp %s/%s .\n" % (myMokkaDir, self.db_dump_name))
    script.write("fi\n")

    # Do something with the additional environment variables
    add_env = self.ops.getOptionsDict("/AvailableTarBalls/%s/%s/%s/AdditionalEnvVar" % (self.platform,
                                                                                        "mokka",
                                                                                        self.applicationVersion))
    if add_env['OK']:
      for key in add_env['Value'].keys():
        script.write('declare -x %s=%s/%s\n' % (key, mySoftwareRoot, add_env['Value'][key]))
    else:
      LOG.verbose("No additional environment variables needed for this application")
    script.close()
    #os.chmod(env_script_name, 0755)
    return S_OK(os.path.abspath(env_script_name))

  #############################################################################

  def determineDBSlice(self):
    """Figure out where the dbSlice is located and what it is called dbSlice is module parameter, set directly via workflow.

    if it is not set and we do not use native CVMFS installation, then we assume
    the SQL file should be called CLICMOkkaDB.sql, but if we are using a native
    CVMFS installation, we need to know where to find the default sql file.

    If we are using ILDConfig, then the dbslice is a tarball in a subfolder of the package.
    This must be specified in the CVMFSDBSlice CS variable
    """
    if self.dbSlice and os.path.exists(self.dbSlice):
      LOG.notice("DBSlice %s already in working directory" % self.dbSlice)
      return S_OK()
    else:
      appVersion = self.step_commons['applicationVersion']
      csPathApplication = "/AvailableTarBalls/%s/%s/%s/" % (self.platform, 'mokka', appVersion)

      # check if ildConfig is used, and then get the dbSlice from there, if it is on cvmfs
      # if the tarball is not on cvmfs we exit, because it will be taken from the ILDConfig tarball
      if 'ILDConfigPackage' in self.workflow_commons:
        LOG.notice("Using ILDConfig from CVMFS, get DBSlice from there")
        config_dir = self.workflow_commons['ILDConfigPackage']
        ildConfigVersion = config_dir.replace("ILDConfig", "")
        ildconfigCSPathApplication = "/AvailableTarBalls/%s/%s/%s/" % (self.platform, 'ildconfig', ildConfigVersion)
        cvmfsDBSlice = Operations().getValue(ildconfigCSPathApplication + "/CVMFSDBSlice")
        cvmfsPath = Operations().getValue(ildconfigCSPathApplication + "/CVMFSPath")
        if cvmfsDBSlice and os.path.exists(cvmfsDBSlice):
          LOG.notice("Getting this dbSlice %s from CVMFS" % cvmfsDBSlice)
          self.untarDBSlice(cvmfsDBSlice)
          return S_OK()
        if cvmfsPath and not cvmfsDBSlice:
          LOG.error("CVMFSDBSlice parameter not set for ILDConfig version", ildConfigVersion)
          return S_ERROR("CVMFSDBSlice parameter not set")
        return S_OK()

      # check if this mokka version is CVMFS native, if not we assume the steeringFile Tarballs has the db slice
      cvmfsPath = Operations().getValue(csPathApplication + "/CVMFSPath")
      if not cvmfsPath:
        LOG.info("This mokka version is not native to CVMFS")
        LOG.info("Assuming CLICMokkaDB.sql")
        self.dbSlice = "CLICMokkaDB.sql"
        return S_OK()

      # Now we know we are CVMFS native, so we have to know CVMFSDBSlice
      cvmfsDBSlice = Operations().getValue(csPathApplication + "/CVMFSDBSlice")
      if not cvmfsDBSlice:
        return S_ERROR("CVMFSDBSlice not defined for mokka version %s " % appVersion)
      LOG.info("Getting this DBSlice: %s" % cvmfsDBSlice)
      # copy the db slice and extract it to local folder?
      # extract the name of the slice from the tarball name
      self.untarDBSlice(cvmfsDBSlice)
      return S_OK()

  def determineRandomSeed(self):
    """determine what the randomSeed should be, depends on production or not."""
    if not self.RandomSeed:
      if self.jobID:
        self.RandomSeed = self.jobID
    if "IS_PROD" in self.workflow_commons:
      self.RandomSeed = int(
          str(
              int(
                  self.workflow_commons["PRODUCTION_ID"]))
          + str(
              int(
                  self.workflow_commons["JOB_ID"])))
    return self.RandomSeed

  def untarDBSlice(self, cvmfsDBSlice):
    """extract the sql slice from some tgz file."""
    dbSliceFileName = cvmfsDBSlice.split("/")[-1]
    if dbSliceFileName[-4:] == ".tgz":
      self.dbSlice = dbSliceFileName[:-4]  # cutaway the ".tgz"
      LOG.notice("untarDBSlice: Using this db %s" % self.dbSlice)
      import tarfile
      dbsliceTar = tarfile.open(cvmfsDBSlice, mode="r:gz")
      dbsliceTar.extractall(path='./')
    else:
      self.dbSlice = dbSliceFileName
    return None



"""Helper functions."""


def find_between(s, first, last):
  """Extract value."""
  try:
    start = s.index(first) + len(first)
    end = s.index(last, start)
    return s[start:end]
  except ValueError:
    return ''



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Remove input data (used in case of merging)"""

from __future__ import absolute_import
from ILCDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC import S_OK, S_ERROR, gLogger

__RCSID__ = '$Id$'
LOG = gLogger.getSubLogger(__name__)


class RemoveInputData(ModuleBase):
  """Remove the input data: to be used when Merging things."""

  def __init__(self):
    super(RemoveInputData, self).__init__()
    self.datMan = DataManager()
    self.enable = True

  def applicationSpecificInputs(self):
    self.enable = self.step_commons.get('Enable', self.enable)
    if not isinstance(self.enable, bool):
      LOG.warn('Enable flag set to non-boolean value %s, setting to False' % self.enable)
      self.enable = False
    return S_OK('Parameters resolved')

  def execute(self):
    """Remove the input data, and pass by failover in case of failure."""
    self.result = self.resolveInputVariables()
    if not self.workflowStatus['OK'] or not self.stepStatus['OK']:
      LOG.verbose('Workflow status = %s, step status = %s' % (self.workflowStatus['OK'], self.stepStatus['OK']))
      return S_OK('No removal of input data attempted')
    if not self.enable:
      LOG.info("Would have tried to remove %s" % self.InputData)
      return S_OK('Input Data Removed')
    try:
      # Try to remove the file list with failover if necessary
      failover = []
      LOG.info('Attempting rm.removeFile("%s")' % (self.InputData))
      result = self.datMan.removeFile(self.InputData)
      LOG.verbose(result)
      if not result['OK']:
        LOG.error('Could not remove files with message:\n"%s"\n\
        Will set removal requests just in case.' % (result['Message']))
        failover = self.InputData
      try:
        if result['Value']['Failed']:
          failureDict = result['Value']['Failed']
          if failureDict:
            LOG.info('Not all files were successfully removed, see "LFN : reason" below\n%s' % (failureDict))
          failover = list(failureDict.keys())
      except KeyError:
        LOG.error('Setting files for removal request to be the input data: %s' % self.InputData)
        failover = self.InputData

      for lfn in failover:
        self.__setFileRemovalRequest(lfn)

      return S_OK('Input Data Removed')
    except OSError as e:
      LOG.exception(e)
      return S_ERROR(e)

    return S_OK()

  def __setFileRemovalRequest(self, lfn):
    """Sets a removal request for a file including all replicas."""
    LOG.info('Setting file removal request for %s' % lfn)
    self.addRemovalRequests([lfn])
